import pytest

from aiidalab_widgets_base import nodes


@pytest.mark.usefixtures("aiida_profile_clean")
def test_nodes_tree_widget(multiply_add_completed_workchain):
    """Test ProcessNodesTreeWidget with a simple `WorkChainNode`"""

    process = multiply_add_completed_workchain
    tree = nodes.NodesTreeWidget()
    tree.nodes = [process]

    # main node is selected
    assert tree.find_node(process.pk).selected is False

    # test the tree node with pk -999 is not in the tree, and raises KeyError
    with pytest.raises(KeyError):
        tree.find_node(-999)

    # the process has two descendants, they finished successfully
    for descendant_process in process.called_descendants:
        tree_node = tree.find_node(descendant_process.pk)
        assert descendant_process.is_finished_ok
        assert tree_node.icon_style == "success"


@pytest.mark.usefixtures("aiida_profile_clean")
def test_open_aiida_node_in_app_widget(multiply_add_completed_workchain):
    """ "Test OpenAiidaNodeInAppWidget with a simple `WorkChainNode`"""

    process = multiply_add_completed_workchain
    open_node_in_app = nodes.OpenAiidaNodeInAppWidget()

    assert len(open_node_in_app.tab.children) == 0

    open_node_in_app.node = process

    assert len(open_node_in_app.tab.children) > 0
    assert open_node_in_app.tab._titles == {
        "0": "Geometry Optimization",
        "1": "Geometry analysis",
        "2": "Isotherm",
    }



import ase
import pytest

import aiidalab_widgets_base as awb
from aiidalab_widgets_base.databases import ComputationalResourcesDatabaseWidget


# timeout for the test
@pytest.mark.timeout(30)
def test_cod_query_widget():
    """Test the COD query widget."""

    widget = awb.CodQueryWidget()

    # Enter the query string.
    widget.inp_elements.value = "Ni Ti"

    # Run the query.
    widget._on_click_query()

    # Select one of the results.
    # TODO: Select a different structure to get rid of the ASE warning:
    # "ase/io/cif.py:401: UserWarning: crystal system 'cubic' is not interpreted
    # for space group 'Pm-3m'. This may result in wrong setting!"
    widget.drop_structure.label = "NiTi (id: 1100132)"

    # Check that the structure was loaded.
    assert isinstance(widget.structure, ase.Atoms)
    assert widget.structure.get_chemical_formula() == "NiTi"


def test_optimade_query_widget():
    """Test the OPTIMADE query widget."""

    widget = awb.OptimadeQueryWidget()

    # At the present state I cannot check much. Most of the variables are locals of the  __init__ method.

    assert widget.structure is None


def test_computational_resources_database_widget():
    """Test the structure browser widget."""
    # Initiate the widget with no arguments.
    widget = ComputationalResourcesDatabaseWidget()
    assert "daint.cscs.ch" in widget.database

    # Initialize the widget with default_calc_job_plugin="cp2k"
    # Note: after migrate to the new database with schema fixed, this test should go with
    # the local defined database rather than relying on the remote one.
    # Same for the quick setup widget.
    widget = ComputationalResourcesDatabaseWidget(default_calc_job_plugin="cp2k")
    assert (
        "merlin7.psi.ch" not in widget.database
    )  # Merlin does not have CP2K installed.

    # Select computer/code
    widget.domain_selector.value = "daint.cscs.ch"
    widget.computer_selector.value = "mc"

    # check the code is not selected
    assert widget.code_selector.value is None

    widget.code_selector.value = "cp2k-9.1"

    # Check that the configuration is provided.
    assert "label" in widget.computer_setup
    assert "hostname" in widget.computer_configure
    assert "filepath_executable" in widget.code_setup

    # test after computer re-select to another, the code selector is reset
    widget.computer_selector.value = "gpu"
    assert widget.code_selector.value is None

    # Simulate reset.
    widget.reset()

    assert widget.computer_setup == {}
    assert widget.computer_configure == {}
    assert widget.code_setup == {}

    # after reset, the computer/code selector is reset
    assert widget.computer_selector.options == ()
    assert widget.code_selector.options == ()
    assert widget.computer_selector.value is None
    assert widget.code_selector.value is None

    # after reset, the domain selector value is reset, but the options are not
    assert widget.domain_selector.value is None
    assert len(widget.domain_selector.options) > 0


def test_resource_database_widget_recognize_template_entry_points():
    """Test that the template like entry points are recognized."""
    # Initiate the widget with no arguments.
    widget = ComputationalResourcesDatabaseWidget()
    assert "daint.cscs.ch" in widget.database

    # Initialize the widget with default_calc_job_plugin="quantumespresso.pw"
    # In merlin, there is a template entry point for Quantum ESPRESSO.
    # Note: after migrate to the new database with schema fixed, this test should go with
    # the local defined database rather than relying on the remote one.
    # Same for the quick setup widget.
    widget = ComputationalResourcesDatabaseWidget(
        default_calc_job_plugin="quantumespresso.pw"
    )
    assert "merlin7.psi.ch" in widget.database

    widget = ComputationalResourcesDatabaseWidget(
        default_calc_job_plugin="quantumespresso.ph"
    )
    assert "merlin7.psi.ch" in widget.database



class ListOrTuppleError(TypeError):
    """Raised when the provided value is not a list or a tupple."""

    def __init__(self, value):
        super().__init__(
            f"The provided value {value!r} is not a list or a tupple, but a {type(value)}."
        )



import ipywidgets as ipw


class LoadingWidget(ipw.HBox):
    """Widget for displaying a loading spinner."""

    def __init__(self, message="Loading", **kwargs):
        self.message = ipw.Label(message)
        super().__init__(
            children=[
                ipw.Label(message),
                ipw.HTML(
                    value="<i class='fa fa-spinner fa-spin fa-2x fa-fw'/>",
                    layout=ipw.Layout(margin="12px 0 6px"),
                ),
            ],
            layout=ipw.Layout(
                justify_content="center",
                align_items="center",
                **kwargs.pop("layout", {}),
            ),
            **kwargs,
        )
        self.add_class("loading")



"""Reusable widgets for AiiDAlab applications."""

_WARNING_TEMPLATE = """
<div style="background-color: #f7f7f7; border: 2px solid #e0e0e0; padding: 20px; border-radius: 5px;">
    <p style="font-size: 16px; font-weight: bold; color: #ff5733;">Warning:</p>
    <p style="font-size: 14px;">The default profile '<span style="font-style: italic;">{profile}</span>' was loaded automatically. This behavior will be removed in the <span style="font-style: italic;">{version}</span>. Please load the profile manually before loading modules from aiidalab-widgets-base by adding the following code at the beginning cell of the notebook:</p>
    <pre style="background-color: #f0f0f0; padding: 10px; border: 1px solid #ccc; font-family: 'Courier New', monospace;">
from aiida import load_profile
load_profile();</pre>
</div>
"""


def load_default_profile():
    """Loads the default profile if none loaded and warn of deprecation."""
    from aiida import load_profile

    load_profile()

    profile = get_profile()
    assert profile is not None, "Failed to load the default profile"

    # raise a deprecation warning
    warning = HTML(_WARNING_TEMPLATE.format(profile=profile.name, version="v3.0.0"))
    display(warning)


# We only detect profile and throw a warning if it is on the notebook
# It is not necessary to do this in the unit tests
def is_running_in_jupyter():
    try:
        from IPython import get_ipython

        if get_ipython() is not None:
            return True
        else:
            return False
    except NameError:
        return False


if is_running_in_jupyter():
    from pathlib import Path

    from aiida.manage import get_profile
    from IPython.display import HTML, display

    # load the default profile if no profile is loaded, and raise a deprecation warning
    # this is a temporary solution to avoid breaking existing notebooks
    # this will be removed in the next major release
    if get_profile() is None:
        load_default_profile()

    from .utils.loaders import load_css

    load_css(css_path=Path(__file__).parent / "static/styles")


from .computational_resources import (
    ComputationalResourcesWidget,
    ComputerDropdownWidget,
)
from .databases import CodQueryWidget, OptimadeQueryWidget
from .elns import ElnConfigureWidget, ElnExportWidget, ElnImportWidget
from .export import ExportButtonWidget
from .loaders import LoadingWidget
from .nodes import NodesTreeWidget, OpenAiidaNodeInAppWidget
from .process import (
    ProcessCallStackWidget,
    ProcessFollowerWidget,
    ProcessInputsWidget,
    ProcessListWidget,
    ProcessMonitor,
    ProcessNodesTreeWidget,
    ProcessOutputsWidget,
    ProcessReportWidget,
    ProgressBarWidget,
    RunningCalcJobOutputWidget,
    SubmitButtonWidget,
)
from .structures import (
    BasicCellEditor,
    BasicStructureEditor,
    SmilesWidget,
    StructureBrowserWidget,
    StructureExamplesWidget,
    StructureManagerWidget,
    StructureUploadWidget,
)
from .viewers import AiidaNodeViewWidget, register_viewer_widget, viewer
from .wizard import WizardAppWidget, WizardAppWidgetStep

__all__ = [
    "AiidaNodeViewWidget",
    "BasicCellEditor",
    "BasicStructureEditor",
    "CodQueryWidget",
    "CodeDatabaseWidget",
    "CodeDropdown",
    "ComputationalResourcesWidget",
    "ComputerDatabaseWidget",
    "ComputerDropdownWidget",
    "ElnConfigureWidget",
    "ElnExportWidget",
    "ElnImportWidget",
    "ExportButtonWidget",
    "LoadingWidget",
    "MultiStructureUploadWidget",
    "NodesTreeWidget",
    "OpenAiidaNodeInAppWidget",
    "OptimadeQueryWidget",
    "ProcessCallStackWidget",
    "ProcessFollowerWidget",
    "ProcessInputsWidget",
    "ProcessListWidget",
    "ProcessMonitor",
    "ProcessNodesTreeWidget",
    "ProcessOutputsWidget",
    "ProcessReportWidget",
    "ProgressBarWidget",
    "RunningCalcJobOutputWidget",
    "SmilesWidget",
    "StructureBrowserWidget",
    "StructureExamplesWidget",
    "StructureManagerWidget",
    "StructureUploadWidget",
    "SubmitButtonWidget",
    "WizardAppWidget",
    "WizardAppWidgetStep",
    "register_viewer_widget",
    "viewer",
]

__version__ = "2.5.0"



import json
from pathlib import Path

import ipywidgets as ipw
import traitlets as tl
from aiida import orm
from IPython.display import clear_output, display

ELN_CONFIG = Path.home() / ".aiidalab" / "aiidalab-eln-config.json"
ELN_CONFIG.parent.mkdir(
    parents=True, exist_ok=True
)  # making sure that the folder exists.


def connect_to_eln(eln_instance=None, **kwargs):
    try:
        from aiidalab_eln import get_eln_connector
    except ImportError:
        return (
            None,
            "AiiDAlab-ELN connector not installed. Install with `pip install aiidalab-eln`",
        )
    # assuming that the connection can only be established to the ELNs
    # with the stored configuration.
    try:
        with open(ELN_CONFIG) as file:
            config = json.load(file)
    except (FileNotFoundError, json.JSONDecodeError, KeyError):
        return (
            None,
            f"Can't open {ELN_CONFIG!r} (ELN configuration file). Instance: {eln_instance}",
        )

    # If no ELN instance was specified, trying the default one.
    if not eln_instance:
        eln_instance = config.pop("default", None)

    if eln_instance:  # The ELN instance could be identified.
        if eln_instance in config:
            eln_config = config[eln_instance]
            eln_type = eln_config.pop("eln_type", None)
        else:  # The selected instance is not present in the config.
            return None, f"Didn't find configuration for the {eln_instance!r} instance."

        # If the ELN type cannot be identified - aborting.
        if not eln_type:
            return None, f"Can't identify the type of {eln_instance} ELN."

        # Everything is alright, can populate the ELN connector
        # with the required info.
        try:
            eln = get_eln_connector(eln_type)(
                eln_instance=eln_instance, **eln_config, **kwargs
            )
        except NotImplementedError as err:
            return None, str(err)
        eln.connect()
        return eln, None

    return (
        None,
        "No ELN instance was provided, the default ELN instance is not configured either. Set a default ELN or select an ELN instance.",
    )


class ElnImportWidget(ipw.VBox):
    node = tl.Instance(orm.Node, allow_none=True)

    def __init__(self, path_to_root="../", **kwargs):
        import requests_cache

        # Used to output additional settings.
        self._output = ipw.Output()

        # Communicate to the user if something isn't right.
        error_message = ipw.HTML()
        super().__init__(children=[error_message], **kwargs)

        eln, msg = connect_to_eln(**kwargs)

        if eln is None:
            url = f"{path_to_root}aiidalab-widgets-base/notebooks/eln_configure.ipynb"
            error_message.value = f"""Warning! The access to ELN is not configured. Please follow <a href={url!r} target="_blank">the link</a> to configure it.</br> More details: {msg!r}"""
            return

        tl.dlink((eln, "node"), (self, "node"))
        # Since the requests cache is enabled globally in aiidalab package, we disable it here to get correct results.
        # This can be removed once https://github.com/aiidalab/aiidalab/issues/196 is fixed.
        with requests_cache.disabled():
            eln.import_data()


class ElnExportWidget(ipw.VBox):
    node = tl.Instance(orm.Node, allow_none=True)

    def __init__(self, path_to_root="../", **kwargs):
        self.path_to_root = path_to_root

        # Send to ELN button.
        send_button = ipw.Button(description="Send to ELN")
        send_button.on_click(self.send_to_eln)

        # Use non-default destination.
        self.modify_settings = ipw.Checkbox(
            description="Update destination.", indent=False
        )
        self.modify_settings.observe(self.handle_output, "value")

        # Used to output additional settings.
        self._output = ipw.Output()

        # Communicate to the user if something isn't right.
        self.message = ipw.HTML()

        children = [
            ipw.HBox([send_button, self.modify_settings]),
            self._output,
            self.message,
        ]
        self.eln, msg = connect_to_eln()
        if self.eln:
            tl.dlink((self, "node"), (self.eln, "node"))
        else:
            self.modify_settings.disabled = True
            send_button.disabled = True
            self.message.value = f"""Warning! The access to ELN is not configured. Please follow <a href="{self.path_to_root}/aiidalab-widgets-base/notebooks/eln_configure.ipynb" target="_blank">the link</a> to configure it.</br> </br> More details: {msg}"""

        super().__init__(children=children, **kwargs)

    @tl.observe("node")
    def _observe_node(self, _=None):
        if self.node is None or self.eln is None:
            return

        info = self.node.base.extras.get("eln", {})
        if not info:
            try:
                q = orm.QueryBuilder().append(
                    orm.Node,
                    filters={"extras": {"has_key": "eln"}},
                    tag="source_node",
                    project="extras.eln",
                )
                q.append(
                    orm.Node,
                    filters={"uuid": self.node.uuid},
                    with_ancestors="source_node",
                )
                info = q.all(flat=True)[0]
            except IndexError:
                info = {}

        self.eln.set_sample_config(**info)

    def send_to_eln(self, _=None):
        import requests_cache

        if self.eln and self.eln.is_connected:
            self.message.value = f"\u29d7 Sending data to {self.eln.eln_instance}..."
            # Since the requests cache is enabled globally in aiidalab package, we disable it here to get correct results.
            # This can be removed once https://github.com/aiidalab/aiidalab/issues/196 is fixed.
            with requests_cache.disabled():
                self.eln.export_data()
            self.message.value = (
                f"\u2705 The data were successfully sent to {self.eln.eln_instance}."
            )
        else:
            self.message.value = f"""\u274c Something isn't right! We were not able to send the data to the "<strong>{self.eln.eln_instance}</strong>" ELN instance. Please follow <a href="{self.path_to_root}/aiidalab-widgets-base/notebooks/eln_configure.ipynb" target="_blank">the link</a> to update the ELN's configuration."""

    def handle_output(self, _=None):
        with self._output:
            clear_output()
            if self.modify_settings.value:
                display(
                    ipw.HTML(
                        f"""Currently used ELN is: "<strong>{self.eln.eln_instance}</strong>". To change it, please follow <a href="{self.path_to_root}/aiidalab-widgets-base/notebooks/eln_configure.ipynb" target="_blank">the link</a>."""
                    )
                )
                display(self.eln.sample_config_editor())


class ElnConfigureWidget(ipw.VBox):
    def __init__(self, **kwargs):
        self._output = ipw.Output()
        self.eln = None

        self.eln_instance = ipw.Dropdown(
            description="ELN:",
            options=("Set up new ELN", {}),
            style={"description_width": "initial"},
        )
        self.update_list_of_elns()
        self.eln_instance.observe(self.display_eln_config, names=["value", "options"])

        self.eln_types = ipw.Dropdown(
            description="ELN type:",
            options=["cheminfo", "openbis"],
            value="cheminfo",
            style={"description_width": "initial"},
        )
        self.eln_types.observe(self.display_eln_config, names=["value", "options"])

        # Buttons.

        # Make current ELN the default.
        default_button = ipw.Button(description="Set as default", button_style="info")
        default_button.on_click(self.set_current_eln_as_default)

        # Save current ELN configuration.
        save_config = ipw.Button(
            description="Save configuration", button_style="success"
        )
        save_config.on_click(self.save_eln_configuration)

        # Erase current ELN from the configuration.
        erase_config = ipw.Button(
            description="Erase configuration", button_style="danger"
        )
        erase_config.on_click(self.erase_current_eln_from_configuration)

        # Check if connection to the current ELN can be established.
        check_connection = ipw.Button(
            description="Check connection", button_style="warning"
        )
        check_connection.on_click(self.check_connection)

        self.my_output = ipw.HTML()

        self.display_eln_config()

        super().__init__(
            children=[
                self.eln_instance,
                self.eln_types,
                self._output,
                ipw.HBox([default_button, save_config, erase_config, check_connection]),
                self.my_output,
            ],
            **kwargs,
        )

    def write_to_config(self, config):
        with open(ELN_CONFIG, "w") as file:
            json.dump(config, file, indent=4)

    def get_config(self):
        try:
            with open(ELN_CONFIG) as file:
                return json.load(file)
        except (FileNotFoundError, json.JSONDecodeError, KeyError):
            return {}

    def update_list_of_elns(self):
        config = self.get_config()
        default_eln = config.pop("default", None)
        if (
            default_eln not in config
        ):  # Erase the default ELN if it is not present in the config
            self.write_to_config(config)
            default_eln = None

        self.eln_instance.options = [("Setup new ELN", {}), *list(config.items())]
        if default_eln:
            self.eln_instance.label = default_eln

    def set_current_eln_as_default(self, _=None):
        self.update_eln_configuration("default", self.eln_instance.label)

    def update_eln_configuration(self, eln_instance, eln_config):
        config = self.get_config()
        config[eln_instance] = eln_config
        self.write_to_config(config)

    def erase_current_eln_from_configuration(self, _=None):
        config = self.get_config()
        config.pop(self.eln_instance.label, None)
        self.write_to_config(config)
        self.update_list_of_elns()

    def check_connection(self, _=None):
        if self.eln:
            err_message = self.eln.connect()
            if self.eln.is_connected:
                self.my_output.value = "\u2705 Connected."
                return
        self.my_output.value = f"\u274c Not connected. {err_message}"

    def display_eln_config(self, value=None):
        """Display ELN configuration specific to the selected type of ELN."""
        try:
            from aiidalab_eln import get_eln_connector
        except ImportError:
            with self._output:
                clear_output()
                msg = "AiiDAlab-ELN connector not installed. Install with `pip install aiidalab-eln`"
                display(ipw.HTML(f"&#10060; {msg}"))
            return

        try:
            eln_class = get_eln_connector(self.eln_types.value)
        except NotImplementedError as err:
            with self._output:
                clear_output()
                display(ipw.HTML("&#10060;" + str(err)))
                return

        self.eln = eln_class(
            eln_instance=self.eln_instance.label if self.eln_instance.value else "",
            **self.eln_instance.value,
        )

        if self.eln_instance.value:
            self.eln_types.value = self.eln.eln_type
            self.eln_types.disabled = True
        else:
            self.eln_types.disabled = False

        with self._output:
            clear_output()
            display(self.eln)

    def save_eln_configuration(self, _=None):
        config = self.eln.get_config()
        eln_instance = config.pop("eln_instance")
        if eln_instance:
            self.update_eln_configuration(eln_instance, config)
            self.update_list_of_elns()



from __future__ import annotations

from pathlib import Path

from IPython.display import Javascript, display


def load_css(css_path: Path | str) -> None:
    """Load and inject CSS stylesheets into the DOM.

    Parameters
    ----------
    `css_path` : `Path` | `str`
        The path to the CSS stylesheet. If the path is a directory,
        all CSS files in the directory will be loaded.
    """
    path = Path(css_path)

    if not path.exists():
        raise FileNotFoundError(f"CSS file or directory not found: {path}")

    filenames = [*path.glob("*.css")] if path.is_dir() else [path]

    for fn in filenames:
        stylesheet = fn.read_text()
        display(
            Javascript(f"""
                var style = document.createElement('style');
                style.type = 'text/css';
                style.innerHTML = `{stylesheet}`;
                document.head.appendChild(style);
            """)
        )



import pytest
from aiida import engine, orm
from aiida.workflows.arithmetic.multiply_add import MultiplyAddWorkChain

import aiidalab_widgets_base as awb


@pytest.mark.usefixtures("aiida_profile_clean")
def test_submit_button_widget(multiply_add_process_builder_ready):
    """Test SubmitButtonWidget with a simple `WorkChainNode`"""

    def hook(_=None):
        pass

    def return_inputs():
        return multiply_add_process_builder_ready

    widget = awb.SubmitButtonWidget(
        process_class=MultiplyAddWorkChain, inputs_generator=return_inputs
    )

    widget.on_submitted(hook)

    assert widget.process is None  # The process is not yet submitted.

    # Simulate the click on the button.
    widget.on_btn_submit_press()
    assert widget.process is not None
    assert isinstance(widget.process, orm.WorkChainNode)


@pytest.mark.usefixtures("aiida_profile_clean")
def test_process_inputs_widget(generate_calc_job_node):
    """Test ProcessInputWidget with a simple `CalcJobNode`"""

    process = generate_calc_job_node(
        inputs={
            "parameters": orm.Int(1),
            "nested": {
                "inner": orm.Int(2),
            },
        }
    )

    # Test the widget can be instantiated with empty inputs
    process_input_widget = awb.ProcessInputsWidget()

    process_input_widget = awb.ProcessInputsWidget(process=process)
    input_dropdown = process_input_widget._inputs

    assert "parameters" in [key for key, _ in input_dropdown.options]
    assert "nested.inner" in [key for key, _ in input_dropdown.options]

    # select the nested input from dropdown and check that the value is displayed)
    uuid = dict(input_dropdown.options)["nested.inner"]
    input_dropdown.value = uuid

    selected_input = orm.load_node(uuid)

    assert process_input_widget.info.value == f"PK: {selected_input.pk}"


@pytest.mark.usefixtures("aiida_profile_clean")
def test_process_outputs_widget(multiply_add_completed_workchain):
    """Test ProcessOutputWidget with a simple `WorkChainNode`"""
    # Test the widget can be instantiated with empty inputs
    widget = awb.ProcessOutputsWidget()

    # Test the widget can be instantiated with a process
    widget = awb.ProcessOutputsWidget(process=multiply_add_completed_workchain)

    # Simulate output selection.
    widget.show_selected_output(change={"new": "result"})


@pytest.mark.usefixtures("aiida_profile_clean")
def test_process_follower_widget(multiply_add_process_builder_ready, daemon_client):
    """Test ProcessFollowerWidget with a simple `WorkChainNode`"""
    # Test the widget can be instantiated with empty inputs
    widget = awb.ProcessFollowerWidget()

    if daemon_client.is_daemon_running:
        daemon_client.stop_daemon(wait=True)
    process = engine.submit(multiply_add_process_builder_ready)

    # Test the widget can be instantiated with a process
    widget = awb.ProcessFollowerWidget(process=process)

    daemon_client.start_daemon()

    # Follow the process till it is completed.
    widget.follow()

    daemon_client.stop_daemon(wait=True)


@pytest.mark.usefixtures("aiida_profile_clean")
def test_process_report_widget(
    multiply_add_process_builder_ready,
    stopped_daemon_client,
    await_for_process_completeness,
):
    """Test ProcessReportWidget with a simple `WorkChainNode`"""
    # Test the widget can be instantiated with empty inputs
    awb.ProcessReportWidget()

    process = engine.submit(multiply_add_process_builder_ready)

    # Test the widget can be instantiated with a process
    widget = awb.ProcessReportWidget(process=process)
    assert (
        widget.value == "No log messages recorded for this entry"
    )  # No report produced yet.

    # Starting the daemon and waiting for the process to complete.
    stopped_daemon_client.start_daemon()
    await_for_process_completeness(process)

    widget.update()
    stopped_daemon_client.stop_daemon(wait=True)


@pytest.mark.usefixtures("aiida_profile_clean")
def test_process_call_stack_widget(
    multiply_add_process_builder_ready, daemon_client, await_for_process_completeness
):
    """Test ProcessCallStackWidget with a simple `WorkChainNode`"""
    from aiidalab_widgets_base.process import ProcessCallStackWidget

    # Test the widget can be instantiated with empty inputs
    ProcessCallStackWidget()

    # Stopping the daemon and submitting the process.
    if daemon_client.is_daemon_running:
        daemon_client.stop_daemon(wait=True)
    process = engine.submit(multiply_add_process_builder_ready)

    # Test the widget can be instantiated with a process
    widget = ProcessCallStackWidget(process=process)
    assert widget.value.endswith("Created")

    # Starting the daemon and waiting for the process to complete.
    daemon_client.start_daemon()
    await_for_process_completeness(process)

    widget.update()
    assert "ArithmeticAddCalculation" in widget.value
    daemon_client.stop_daemon(wait=True)


@pytest.mark.usefixtures("aiida_profile_clean")
def test_progress_bar_widget(
    multiply_add_process_builder_ready, daemon_client, await_for_process_completeness
):
    """Test ProgressBarWidget with a simple `WorkChainNode`"""
    from aiidalab_widgets_base import ProgressBarWidget

    # Test the widget can be instantiated with empty inputs
    ProgressBarWidget()

    # Stopping the daemon and submitting the process.
    if daemon_client.is_daemon_running:
        daemon_client.stop_daemon(wait=True)
    process = engine.submit(multiply_add_process_builder_ready)

    # Test the widget can be instantiated with a process
    widget = ProgressBarWidget(process=process)
    assert widget.state.value == "Created"

    # Starting the daemon and waiting for the process to complete.
    daemon_client.start_daemon()
    await_for_process_completeness(process)

    widget.update()
    assert widget.state.value == "Finished"
    daemon_client.stop_daemon(wait=True)


@pytest.mark.usefixtures("aiida_profile_clean")
def test_calcjob_output_widget(generate_calc_job_node):
    """Test CalcJobOutputWidget with a simple `WorkChainNode`"""
    from aiidalab_widgets_base.process import CalcJobOutputWidget

    process = generate_calc_job_node(
        inputs={
            "parameters": orm.Int(1),
            "nested": {
                "inner": orm.Int(2),
            },
        }
    )
    # Test the widget can be instantiated with empty inputs
    CalcJobOutputWidget()

    # Test the widget can be instantiated with a process
    CalcJobOutputWidget(calculation=process)


@pytest.mark.usefixtures("aiida_profile_clean")
def test_running_calcjob_output_widget(generate_calc_job_node):
    """Test RunningCalcJobOutputWidget with a simple `WorkChainNode`"""
    from aiidalab_widgets_base.process import RunningCalcJobOutputWidget

    process = generate_calc_job_node(
        inputs={
            "parameters": orm.Int(1),
            "nested": {
                "inner": orm.Int(2),
            },
        }
    )

    widget = RunningCalcJobOutputWidget()
    widget.process = process


@pytest.mark.usefixtures("aiida_profile_clean")
def test_process_list_widget(multiply_add_completed_workchain):
    """Test ProcessListWidget with a simple `WorkChainNode`"""
    awb.ProcessListWidget()


@pytest.mark.usefixtures("aiida_profile_clean")
def test_process_monitor(
    multiply_add_process_builder_ready, daemon_client, await_for_process_completeness
):
    """Test ProcessMonitor with a simple `WorkChainNode`"""
    awb.ProcessMonitor()

    # Stopping the daemon and submitting the process.
    if daemon_client.is_daemon_running:
        daemon_client.stop_daemon(wait=True)
    process = engine.submit(multiply_add_process_builder_ready)

    test_variable = False

    def f():
        nonlocal test_variable
        test_variable = True

    widget = awb.ProcessMonitor(value=process.uuid, callbacks=[f])

    # Starting the daemon and waiting for the process to complete.
    daemon_client.start_daemon()
    await_for_process_completeness(process)

    widget.join()  # Make sure the thread is finished.

    assert test_variable
    daemon_client.stop_daemon(wait=True)


@pytest.mark.usefixtures("aiida_profile_clean")
def test_process_nodes_tree_widget(multiply_add_completed_workchain):
    """Test ProcessNodesTreeWidget with a simple `WorkChainNode`"""
    awb.ProcessNodesTreeWidget(value=multiply_add_completed_workchain.uuid)



"""The wizard application allows the implication of a Wizard-like GUI.

Authors:

    * Carl Simon Adorf <simon.adorf@epfl.ch>
"""

import enum

import ipywidgets as ipw
import traitlets as tl


class AtLeastTwoStepsWizardError(ValueError):
    """Using WizardAppWidget only makes sense if the number of setps is at least two."""

    def __init__(self, steps):
        super().__init__(
            f"The number of steps of a WizardAppWidget must be at least two, but {len(steps)} were provided."
        )


class WizardAppWidgetStep(tl.HasTraits):
    "One step of a WizardAppWidget."

    class State(enum.Enum):
        """Each step is always in one specific state.

        The state is used to determine:

            1) how the step is visually presented to the user, and
            2) whether the next step is accessible (i.e. reached the SUCCESS state).

        App developers are encouraged to use the step states to couple application
        logic and interface. In general, all widget changes should trigger
        a re-evaluation of the step state, and states also determine whether certain
        widgets are enabled or disabled.

        A step can be in one of the following states:

            INIT: The initial state, usually all widgets disabled.
            READY: The step (widget) is ready for user input (some or all widgets enabled).
            CONFIGURED: The step is in a consistent configuration awaiting confirmation.
            ACTIVE: The step is carrying out a runtime operation.
            SUCCESS: A configuration has been confirmed / a runtime operation successfully finished.
            FAIL: A runtime operation has failed in an unrecoverable way.

        Not all steps must implement all states, for example:

            - the first step does not need an INIT state
            - a step without runtime process should not have an ACTIVE or FAIL state
            - a "review & confirm" step does not require a READY state.
            - a step without configuration options (e.g. pure "review & confirm" step)

        Important: The next step is only accessible if the current step is within the SUCCESS
        state!
        """

        INIT = 0  # the step is initialized and all widgets are typically disabled

        # The step is correctly configured and can in principle be confirmed.
        CONFIGURED = 1  # configuration is valid
        READY = 2  # step is ready for user input
        ACTIVE = 3  # step is carrying out a runtime operation
        SUCCESS = 4  # step has successfully completed

        # All error states have negative codes
        FAIL = -1  # the step has unrecoverably failed

    state = tl.UseEnum(State)
    auto_advance = tl.Bool()

    def can_reset(self):
        return hasattr(self, "reset")


class WizardAppWidget(ipw.VBox):
    ICON_SEPARATOR = "\u2000"  # en-dash  (placed between title and icon)

    ICONS = {
        WizardAppWidgetStep.State.INIT: "\u25cb",
        WizardAppWidgetStep.State.READY: "\u25ce",
        WizardAppWidgetStep.State.CONFIGURED: "\u25cf",
        WizardAppWidgetStep.State.ACTIVE: "\u231b",
        WizardAppWidgetStep.State.SUCCESS: "\u2713",
        WizardAppWidgetStep.State.FAIL: "\u00d7",
    }

    selected_index = tl.Int(allow_none=True)

    def __init__(self, steps, show_header=True, **kwargs):
        # The number of steps must be greater than one
        # for this app's logic to make sense.
        if len(steps) < 2:
            raise AtLeastTwoStepsWizardError(steps)

        self.steps = steps

        # Unzip the steps to titles and widgets.
        self.titles, widgets = zip(*steps)

        # Initialize the accordion with the widgets ...
        self.accordion = ipw.Accordion(children=widgets)
        self._update_titles()
        ipw.link((self.accordion, "selected_index"), (self, "selected_index"))

        # Watch for changes to each step's state
        for widget in widgets:
            if not widget.has_trait("state"):
                raise TypeError(
                    f"The provided {widget!r} as wizard app step has no `state` trait. "
                    "It is expected that step classes are derived from the WizardAppWidgetStep class."
                )
            widget.observe(self._update_step_state, names=["state"])

        self.reset_button = ipw.Button(
            description="Reset",
            icon="undo",
            layout=ipw.Layout(width="auto", flex="1 1 auto"),
            tooltip="Reset the app to start over (if possible)",
            disabled=True,
        )
        self.reset_button.on_click(self._on_click_reset_button)

        # Create a back-button, to switch to the previous step when possible:
        self.back_button = ipw.Button(
            description="Previous step",
            icon="step-backward",
            layout=ipw.Layout(width="auto", flex="1 1 auto"),
            tooltip="Go to the previous step.",
            disabled=True,
        )
        self.back_button.on_click(self._on_click_back_button)

        # Create a next-button, to switch to the next step when appropriate:
        self.next_button = ipw.Button(
            description="Next step",
            icon="step-forward",
            layout=ipw.Layout(width="auto", flex="1 1 auto"),
            tooltip="Go to the next step.",
            disabled=True,
        )
        self.next_button.on_click(self._on_click_next_button)

        self.header = ipw.HBox(
            children=[self.back_button, self.reset_button, self.next_button]
        )
        self.show_header = show_header

        super().__init__(children=[self.header, self.accordion], **kwargs)

    @property
    def show_header(self):
        return self.header.layout.display != "none"

    @show_header.setter
    def show_header(self, value):
        self.header.layout.display = "flex" if value else "none"

    def _update_titles(self):
        for i, (title, widget) in enumerate(zip(self.titles, self.accordion.children)):
            icon = self.ICONS.get(widget.state, str(widget.state).upper())
            self.accordion.set_title(i, f"{icon} Step {i + 1}: {title}")

    def _consider_auto_advance(self, _=None):
        """Determine whether the app should automatically advance to the next step.

        This is performed whenever the current step is within the SUCCESS state and has
        the auto_advance attribute set to True.
        """
        if self.accordion.selected_index is None:  # All children are hidden
            return

        with self.hold_trait_notifications():
            index = self.accordion.selected_index
            last_step_selected = index + 1 == len(self.accordion.children)
            selected_widget = self.accordion.children[index]
            if (
                selected_widget.auto_advance
                and not last_step_selected
                and selected_widget.state == WizardAppWidgetStep.State.SUCCESS
            ):
                self.accordion.selected_index += 1

    def _update_step_state(self, _):
        with self.hold_trait_notifications():
            self._update_titles()
            self._update_buttons()
            self._consider_auto_advance()

    @tl.observe("selected_index")
    def _observe_selected_index(self, change):
        "Activate/deactivate the next-button based on which step is selected."
        self._update_buttons()

    def can_reset(self):
        steps = [
            self.accordion.children[idx] for idx in range(len(self.accordion.children))
        ]

        if not all(step.can_reset() for step in steps):
            return False

        if any(step.state is not WizardAppWidgetStep.State.INIT for step in steps):
            return True

        return False

    def _update_buttons(self):
        with self.hold_trait_notifications():
            index = self.accordion.selected_index
            if index is None:
                self.back_button.disabled = True
                self.next_button.disabled = True
                self.reset_button.disabled = True
            else:
                first_step_selected = index == 0
                last_step_selected = index + 1 == len(self.accordion.children)
                selected_widget = self.accordion.children[index]

                self.back_button.disabled = (
                    first_step_selected
                    or selected_widget.state
                    in (
                        WizardAppWidgetStep.State.ACTIVE,
                        WizardAppWidgetStep.State.SUCCESS,
                        WizardAppWidgetStep.State.FAIL,
                    )
                )
                self.next_button.disabled = (
                    last_step_selected
                    or selected_widget.state is not WizardAppWidgetStep.State.SUCCESS
                )

                self.reset_button.disabled = not self.can_reset()

    def reset(self, step=0):
        """Reset the app up to the given step.

        For example, with step=0 (the default), the whole app is reset.
        With step=1, all but the first step are reset.
        """
        with self.hold_sync():
            for index in reversed(range(step, len(self.accordion.children))):
                if hasattr(self.accordion.children[index], "reset"):
                    self.accordion.children[index].reset()
            self.accordion.selected_index = step

    def _on_click_reset_button(self, _):
        self.reset()

    def _on_click_back_button(self, _):
        self.accordion.selected_index -= 1

    def _on_click_next_button(self, _):
        self.accordion.selected_index += 1



import ipywidgets as ipw
import traitlets as tl

from aiidalab_widgets_base import WizardAppWidget, WizardAppWidgetStep


def test_wizard_app_widget():
    class Step1(ipw.HBox, WizardAppWidgetStep):
        config = tl.Bool()

        def __init__(self, **kwargs):
            self.order_button = ipw.Button(description="Submit order", disabled=False)
            self.order_button.on_click(self.submit_order)
            super().__init__(children=[self.order_button], **kwargs)

        def submit_order(self, _=None):
            self.config = True

        @tl.default("config")
        def _default_config(self):
            return False

        def reset(self):
            self.config = False

        @tl.observe("config")
        def _observe_config(self, _=None):
            self.state = self.State.SUCCESS if self.config else self.State.INIT

    class Step2(ipw.HBox, WizardAppWidgetStep):
        config = tl.Bool()

        def __init__(self, **kwargs):
            self.results = ipw.HTML("Results")
            super().__init__(children=[self.results], **kwargs)

        def submit_order(self, _=None):
            pass

        def reset(self):
            pass

        @tl.default("config")
        def _default_config(self):
            return False

        @tl.observe("config")
        def _observe_config(self, change):
            if self.config:
                self.state = self.State.READY
            else:
                self.state = self.State.INIT

    s1 = Step1(auto_advance=True)
    s2 = Step2(auto_advance=True)
    tl.dlink((s1, "config"), (s2, "config"))

    widget = WizardAppWidget(
        steps=[
            ("Setup", s1),
            ("View results", s2),
        ],
    )

    # Check initial state.
    assert s1.state == s1.State.INIT
    assert s2.state == s2.State.INIT
    assert widget.accordion.selected_index == 0
    assert widget.next_button.disabled is True
    assert widget.back_button.disabled is True
    assert widget.accordion.get_title(0) == "○ Step 1: Setup"
    assert widget.accordion.get_title(1) == "○ Step 2: View results"
    assert not widget.can_reset()

    # Check state after finishing the first step.
    s1.order_button.click()
    assert s1.state == s1.State.SUCCESS
    assert s2.state == s2.State.READY
    assert widget.accordion.selected_index == 1
    assert widget.next_button.disabled is True
    assert widget.back_button.disabled is False
    assert widget.accordion.get_title(0) == "✓ Step 1: Setup"
    assert widget.accordion.get_title(1) == "◎ Step 2: View results"
    assert widget.can_reset()

    # Check state after resetting the app.
    widget.reset_button.click()
    assert s1.state == s1.State.INIT
    assert s2.state == s2.State.INIT
    assert widget.back_button.disabled is True
    assert widget.next_button.disabled is True
    assert widget.accordion.get_title(0) == "○ Step 1: Setup"
    assert widget.accordion.get_title(1) == "○ Step 2: View results"
    assert not widget.can_reset()

    # Check state after finishing the first step again.
    s1.config = True  # This should trigger an attempt to advance to the next step.
    assert s1.state == s1.State.SUCCESS
    assert s2.state == s2.State.READY
    assert widget.accordion.selected_index == 1
    assert widget.next_button.disabled is True
    assert widget.back_button.disabled is False
    assert widget.accordion.get_title(0) == "✓ Step 1: Setup"
    assert widget.accordion.get_title(1) == "◎ Step 2: View results"
    assert widget.can_reset()

    # Click on the back button.
    widget.back_button.click()
    assert s1.state == s1.State.SUCCESS
    assert s2.state == s2.State.READY
    assert widget.accordion.selected_index == 0
    assert widget.next_button.disabled is False
    assert widget.back_button.disabled is True
    assert widget.can_reset()

    # Click on the next button.
    widget.next_button.click()
    assert widget.accordion.selected_index == 1
    assert widget.next_button.disabled is True
    assert widget.back_button.disabled is False



"""Widgets to work with processes."""

# pylint: disable=no-self-use
# Built-in imports
from __future__ import annotations

import inspect
import os
import sys
import threading
import time
import traceback
import uuid
import warnings

import ipywidgets as ipw
import traitlets as tl
from aiida import engine, orm
from aiida.cmdline.utils.ascii_vis import format_call_graph
from aiida.cmdline.utils.common import (
    get_calcjob_report,
    get_process_function_report,
    get_workchain_report,
)
from aiida.common.exceptions import NotExistentAttributeError
from aiida.tools.query.calculation import CalculationQueryBuilder
from IPython.display import HTML, Javascript, clear_output, display

# Local imports.
from .nodes import NodesTreeWidget
from .utils import exceptions
from .viewers import viewer


def get_running_calcs(process):
    """Takes a process and yeilds running children calculations."""

    # If a process is a running calculation - returning it
    if issubclass(type(process), orm.CalcJobNode) and not process.is_sealed:
        yield process

    # If the process is a running work chain - returning its children
    elif issubclass(type(process), orm.WorkChainNode) and not process.is_sealed:
        for out_link in process.get_outgoing():
            if (
                isinstance(out_link.node, orm.ProcessNode)
                and not out_link.node.is_sealed
            ):
                yield from get_running_calcs(out_link.node)


class SubmitButtonWidget(ipw.VBox):
    """Submit button class that creates submit button jupyter widget."""

    process = tl.Instance(orm.ProcessNode, allow_none=True)

    def __init__(  # pylint: disable=too-many-arguments
        self,
        process_class,
        inputs_generator,
        description="Submit",
        disable_after_submit=True,
        append_output=False,
        **kwargs,
    ):
        """Submit Button widget.

        process_class (Process): Process class to submit.

        inputs_generator (func): Function that returns inputs dictionary or inputs builder.

        description (str): Description written on the submission button.

        disable_after_submit (bool): Whether to disable the button after the process was submitted.

        append_output (bool): Whether to clear widget output for each subsequent submission.
        """

        self.path_to_root = kwargs.get("path_to_root", "../")
        if inspect.isclass(process_class) and issubclass(process_class, engine.Process):
            self._process_class = process_class
        else:
            raise ValueError(
                f"process_class argument must be a sublcass of {engine.Process}, got {process_class}"
            )

        # Checking if the inputs generator is callable
        if callable(inputs_generator):
            self.inputs_generator = inputs_generator
        else:
            raise TypeError(
                "The `inputs_generator` argument must be a function that "
                f"returns input dictionary, got {type(inputs_generator)}"
            )

        self.disable_after_submit = disable_after_submit
        self.append_output = append_output

        self.btn_submit = ipw.Button(description=description, disabled=False)
        self.btn_submit.on_click(self.on_btn_submit_press)
        self.submit_out = ipw.HTML("")
        self._run_after_submitted = []

        super().__init__(children=[self.btn_submit, self.submit_out])

    def on_click(self, function):
        self.btn_submit.on_click(function)

    def on_btn_submit_press(self, _=None):
        """When submit button is pressed."""

        if not self.append_output:
            self.submit_out.value = ""

        inputs = self.inputs_generator()
        if inputs is None:
            if self.append_output:
                self.submit_out.value += (
                    "SubmitButtonWidget: did not recieve the process inputs.<br>"
                )
            else:
                self.submit_out.value = (
                    "SubmitButtonWidget: did not recieve the process inputs."
                )
        else:
            if self.disable_after_submit:
                self.btn_submit.disabled = True
            if isinstance(inputs, engine.ProcessBuilder):
                self.process = engine.submit(inputs)
            else:
                self.process = engine.submit(self._process_class, **inputs)

            if self.append_output:
                self.submit_out.value += f"""Submitted process {self.process}. Click
                <a href={self.path_to_root}aiidalab-widgets-base/notebooks/process.ipynb?id={self.process.pk}
                target="_blank">here</a> to follow.<br>"""
            else:
                self.submit_out.value = f"""Submitted process {self.process}. Click
                <a href={self.path_to_root}aiidalab-widgets-base/notebooks/process.ipynb?id={self.process.pk}
                target="_blank">here</a> to follow."""

            for func in self._run_after_submitted:
                func(self.process)

    def on_submitted(self, function):
        """Run functions after a process has been submitted successfully."""
        self._run_after_submitted.append(function)


class ProcessInputsWidget(ipw.VBox):
    """Widget to select and show process inputs."""

    process = tl.Instance(orm.ProcessNode, allow_none=True)

    def __init__(self, process=None, **kwargs):
        self.process = process
        self.output = ipw.Output()
        self.info = ipw.HTML()

        self.flat_mapping = self.generate_flat_mapping(process=process) or {}
        inputs_list = list(self.flat_mapping.items())

        self._inputs = ipw.Dropdown(
            options=[("Select input", ""), *inputs_list],
            description="Select input:",
            style={"description_width": "initial"},
            disabled=False,
        )
        self._inputs.observe(self.show_selected_input, names=["value"])
        super().__init__(
            children=[ipw.HBox([self._inputs, self.info]), self.output], **kwargs
        )

    def generate_flat_mapping(
        self, process: orm.ProcessNode | None = None
    ) -> None | dict[str, str]:
        """Generate a dict of input to node uuid mapping.

        If the input port is a namespace, it will further parse the namespace and attach the entity the
        `<namespace>.<childnamespace>.<node_name>` format.

        :param process: Process node.
        :return: Dict of flatten embed key name to node UUID."""
        from collections.abc import Mapping

        from aiida.common.links import LinkType

        if process is None:
            return None

        nested_dict = process.base.links.get_incoming(
            link_type=(LinkType.INPUT_CALC, LinkType.INPUT_WORK)
        ).nested()

        def flatten(d, parent_key="", sep="."):
            items = []
            for key, value in d.items():
                new_key = parent_key + sep + key if parent_key else key
                if isinstance(value, Mapping):
                    items.extend(flatten(value, new_key, sep=sep).items())
                else:
                    items.append((new_key, value.uuid))
            return dict(items)

        options_map = flatten(nested_dict)

        return options_map

    def show_selected_input(self, change=None):
        """Function that displays process inputs selected in the `inputs` Dropdown widget."""
        with self.output:
            self.info.value = ""
            clear_output()
            if change["new"]:
                selected_input = orm.load_node(change["new"])
                self.info.value = f"PK: {selected_input.pk}"
                display(viewer(selected_input))


class ProcessOutputsWidget(ipw.VBox):
    """Widget to select and show process outputs."""

    process = tl.Instance(orm.ProcessNode, allow_none=True)

    def __init__(self, process=None, **kwargs):
        self.process = process
        self.output = ipw.Output()
        self.info = ipw.HTML()
        outputs_list = (
            [(output.title(), output) for output in self.process.outputs]
            if self.process
            else []
        )
        outputs = ipw.Dropdown(
            options=[("Select output", ""), *outputs_list],
            label="Select output",
            description="Select outputs:",
            style={"description_width": "initial"},
            disabled=False,
        )
        outputs.observe(self.show_selected_output, names=["value"])
        super().__init__(
            children=[ipw.HBox([outputs, self.info]), self.output], **kwargs
        )

    def show_selected_output(self, change=None):
        """Function that displays process output selected in the `outputs` Dropdown widget."""
        with self.output:
            self.info.value = ""
            clear_output()
            if change["new"]:
                selected_output = self.process.outputs[change["new"]]
                self.info.value = f"PK: {selected_output.pk}"
                display(viewer(selected_output))


class ProcessFollowerWidget(ipw.VBox):
    """A Widget that follows a process until finished."""

    process = tl.Instance(orm.ProcessNode, allow_none=True)

    def __init__(
        self,
        process=None,
        followers=None,
        update_interval=1.0,
        path_to_root="../",
        **kwargs,
    ):
        """Initiate all the followers."""
        self._monitor = None

        self.process = process
        self._run_after_completed = []
        self.update_interval = update_interval
        self.followers = []
        if followers is not None:
            for follower in followers:
                follower.process = self.process
                follower.path_to_root = path_to_root
                self.followers.append(
                    ipw.VBox(
                        [
                            ipw.HTML(f"<h2><b>{follower.title}</b></h2>"),
                            follower,
                        ]
                    )
                )
        self.output = ipw.HTML()
        super().__init__(children=[self.output, *self.followers], **kwargs)
        self.update()

    def update(self):
        for follower in self.followers:
            follower.children[1].update()

    def follow(self, detach=False):
        """Initiate following the process with or without blocking."""
        if self.process is None:
            self.output.value = """<font color="red"> ProcessFollowerWidget: process
            is set to 'None', nothing to follow. </font>"""
            return
        self.output.value = ""

        if self._monitor is None:
            self._monitor = ProcessMonitor(
                callbacks=[self.update],
                on_sealed=self._run_after_completed,
                timeout=self.update_interval,
            )
            ipw.dlink(
                (self, "process"), (self._monitor, "value"), transform=lambda x: x.uuid
            )

        if not detach:
            self._monitor.join()

    def on_completed(self, function):
        """Run functions after a process has been completed."""
        if self._monitor is not None:
            raise exceptions.CantRegisterCallbackError(function)
        self._run_after_completed.append(function)


class ProcessReportWidget(ipw.HTML):
    """Widget that shows process report."""

    process = tl.Instance(orm.ProcessNode, allow_none=True)
    value = tl.Unicode(allow_none=True)

    def __init__(self, title="Process Report", **kwargs):
        self.title = title
        self.max_depth = None
        self.indent_size = 2
        self.levelname = "REPORT"
        super().__init__(**kwargs)
        self.update()

    def update(self):
        """Update report that is shown."""
        if self.process is None:
            return

        if isinstance(self.process, orm.CalcJobNode):
            string = get_calcjob_report(self.process)
        elif isinstance(self.process, orm.WorkChainNode):
            string = get_workchain_report(
                self.process, self.levelname, self.indent_size, self.max_depth
            )
        elif isinstance(self.process, (orm.CalcFunctionNode, orm.WorkFunctionNode)):
            string = get_process_function_report(self.process)
        else:
            string = f"Nothing to show for node type {self.process.__class__}"
        self.value = string.replace("\n", "<br/>")


class ProcessCallStackWidget(ipw.HTML):
    """Widget that shows process call stack."""

    process = tl.Instance(orm.ProcessNode, allow_none=True)

    def __init__(self, title="Process Call Stack", path_to_root="../", **kwargs):
        self.title = title
        self.path_to_root = path_to_root
        super().__init__(**kwargs)
        self.update()

    def update(self):
        """Update the call stack that is shown."""
        if self.process is None:
            return
        string = format_call_graph(self.process, info_fn=self.calc_info)
        self.value = (
            string.replace("\n", "<br/>").replace(" ", "&nbsp;").replace("#space#", " ")
        )

    # The third parameter 'call_link_label', added in AiiDA 2.4, is not used here.
    # https://github.com/aiidateam/aiida-core/pull/6056
    def calc_info(self, node, _=False):
        """Return a string with the summary of the state of a CalculationNode."""

        if not isinstance(node, orm.ProcessNode):
            raise TypeError(f"Unknown type: {type(node)}")

        process_state = node.process_state.value.capitalize()
        pk = f"""<a#space#href={self.path_to_root}aiidalab-widgets-base/notebooks/process.ipynb?id={node.pk}#space#target="_blank">{node.pk}</a>"""

        if node.exit_status is not None:
            string = f"{node.process_label}<{pk}> {process_state} [{node.exit_status}]"
        else:
            string = f"{node.process_label}<{pk}> {process_state}"

        if isinstance(node, orm.WorkChainNode) and node.stepper_state_info:
            string += f" [{node.stepper_state_info}]"
        return string


class ProgressBarWidget(ipw.VBox):
    """A bar showing the proggress of a process."""

    process = tl.Instance(orm.ProcessNode, allow_none=True)

    def __init__(self, title="Progress Bar", **kwargs):
        """Initialize ProgressBarWidget."""

        self.title = title
        self.correspondance = {
            "created": 0,
            "running": 1,
            "waiting": 1,
            "killed": 2,
            "excepted": 2,
            "finished": 2,
        }
        self.progress_bar = ipw.IntProgress(
            value=0,
            min=0,
            max=2,
            description="Progress:",
            bar_style="warning",  # 'success', 'info', 'warning', 'danger' or ''
            orientation="horizontal",
            layout={"width": "800px"},
        )
        self.state = ipw.HTML(
            description="Calculation state:",
            value="Created",
            style={"description_width": "initial"},
        )
        super().__init__(children=[self.progress_bar, self.state], **kwargs)
        self.update()

    def update(self):
        """Update the bar."""
        if self.process is None:
            return
        self.progress_bar.value = self.correspondance[self.current_state]
        if self.current_state == "finished":
            self.progress_bar.bar_style = "success"
        elif self.current_state in ["killed", "excepted"]:
            self.progress_bar.bar_style = "danger"
        else:
            self.progress_bar.bar_style = "info"
        self.state.value = self.current_state.capitalize()

    @property
    def current_state(self):
        return self.process.process_state.value


class CalcJobOutputWidget(ipw.Textarea):
    """Output of a calculation."""

    calculation = tl.Instance(orm.CalcJobNode, allow_none=True)

    def __init__(self, **kwargs):
        default_params = {
            "value": "",
            "placeholder": "Calculation output will appear here",
            "description": "Calculation output:",
            "layout": {"width": "900px", "height": "300px"},
            "disabled": False,
            "style": {"description_width": "initial"},
        }
        default_params.update(kwargs)
        self.output = []

        # Hack to make font monospace. As far as I am aware, currently there are no better ways.
        display(HTML("<style>textarea, input { font-family: monospace; }</style>"))

        super().__init__(**default_params)

    @tl.observe("calculation")
    def _change_calculation(self, _=None):
        """Reset things if the observed calculation has changed."""
        self.output = []
        self.value = ""

    def update(self):
        """Update the displayed output and scroll to its end.

        NOTE: when this widgets is called by ProcessFollowerWidget in non-blocking manner
        the auto-scrolling won't work. There used to be a function for the Textarea widget,
        but it didn't work properly and got removed. For more information please visit:
        https://github.com/jupyter-widgets/ipywidgets/issues/1815"""

        if self.calculation is None:
            return

        try:
            output_file_path = os.path.join(
                self.calculation.outputs.remote_folder.get_remote_path(),
                self.calculation.base.attributes.get("output_filename"),
            )
        except KeyError:
            self.placeholder = (
                "The `output_filename` attribute is not set for "
                f"{self.calculation.process_class}. Nothing to show."
            )
        except NotExistentAttributeError:
            self.placeholder = (
                "The object `remote_folder` was not found among the process outputs. "
                "Nothing to show."
            )
        else:
            if os.path.exists(output_file_path):
                with open(output_file_path) as fobj:
                    difference = fobj.readlines()[
                        len(self.output) : -1
                    ]  # Only adding the difference
                    self.output += difference
                    self.value += "".join(difference)

        # Auto scroll down. Doesn't work in detached mode.
        # Also a hack as it is applied to all the textareas
        display(
            Javascript(
                """
            $('textarea').each(function(){
                // get the id
                var id_value = $(this).attr('id');

                if (typeof id_value !== 'undefined') {
                    // the variable is defined
                    var textarea = document.getElementById(id_value);
                    textarea.scrollTop = textarea.scrollHeight;
                }
            });"""
            )
        )


class RunningCalcJobOutputWidget(ipw.VBox):
    """Show an output of selected running child calculation."""

    process = tl.Instance(orm.ProcessNode, allow_none=True)

    def __init__(self, title="Running Job Output", **kwargs):
        self.title = title
        self.selection = ipw.Dropdown(
            description="Select calculation:",
            options=tuple((p.pk, p) for p in get_running_calcs(self.process)),
            style={"description_width": "initial"},
        )
        self.output = CalcJobOutputWidget()
        super().__init__(children=[self.selection, self.output], **kwargs)
        self.update()

    def update(self):
        """Update the displayed output."""
        if self.process is None:
            return
        with self.hold_trait_notifications():
            old_label = self.selection.label
            self.selection.options = tuple(
                (str(p.pk), p) for p in get_running_calcs(self.process)
            )
            # If the selection remains the same.
            if old_label in self.selection.options:
                self.label = old_label  # After changing options trait, the label and value traits might change as well.
            # If selection has changed
            else:
                self.output.calculation = self.selection.value
        self.output.update()


class ProcessListWidget(ipw.VBox):
    """List of AiiDA processes.

    past_days (int): Sumulations that were submitted in the last `past_days`.

    incoming_node (str): Trait that takes node uuid that must
    be among the input nodes of the process of interest.

    outgoing_node (str): Trait that takes node uuid that must
    be among the output nodes of the process of interest.

    process_states (list): List of allowed process states.

    process_label (str): Show process states of type `process_label`.

    description_contains (str): string that should be present in the description of a process node.

    """

    past_days = tl.Int(7)
    incoming_node = tl.Unicode(allow_none=True)
    outgoing_node = tl.Unicode(allow_none=True)
    process_states = tl.List()
    process_label = tl.Unicode(allow_none=True)
    description_contains = tl.Unicode(allow_none=True)

    def __init__(self, path_to_root="../", **kwargs):
        self.path_to_root = path_to_root
        self.table = ipw.HTML()
        self.output = ipw.HTML()
        update_button = ipw.Button(description="Update now")
        update_button.on_click(self.update)
        super().__init__(
            children=[ipw.HBox([self.output, update_button]), self.table], **kwargs
        )
        self.update()

    def update(self, _=None):
        """Perform the query."""
        import pandas as pd

        pd.set_option("max_colwidth", 40)
        # Here we are defining properties of 'df' class (specified while exporting pandas table into html).
        # Since the exported object is nothing more than HTML table, all 'standard' HTML table settings
        # can be applied to it as well.
        # For more information on how to controle the table appearance please visit:
        # https://css-tricks.com/complete-guide-table-element/
        self.table.value = """
        <style>
            .df { border: none; }
            .df tbody tr:nth-child(odd) { background-color: #e5e7e9; }
            .df tbody tr:nth-child(odd):hover { background-color:   #f5b7b1; }
            .df tbody tr:nth-child(even):hover { background-color:  #f5b7b1; }
            .df tbody td { min-width: 150px; text-align: center; border: none }
            .df th { text-align: center; border: none;  border-bottom: 1px solid black;}
        </style>
        """
        builder = CalculationQueryBuilder()
        filters = builder.get_filters(
            all_entries=False,
            process_state=self.process_states,
            process_label=self.process_label,
            exit_status=None,
            failed=None,
        )
        relationships = {}
        if self.incoming_node:
            relationships = {
                **relationships,
                **{"with_outgoing": orm.load_node(self.incoming_node)},
            }

        if self.outgoing_node:
            relationships = {
                **relationships,
                **{"with_incoming": orm.load_node(self.outgoing_node)},
            }

        query_set = builder.get_query_set(
            filters=filters,
            past_days=None if self.past_days < 0 else self.past_days,
            order_by={"ctime": "desc"},
            relationships=relationships,
        )
        projected = builder.get_projected(
            query_set,
            projections=[
                "pk",
                "ctime",
                "process_label",
                "state",
                "process_status",
                "description",
            ],
        )
        dataf = pd.DataFrame(projected[1:], columns=projected[0])

        # Keep only process that contain the requested string in the description.
        if self.description_contains:
            dataf = dataf[dataf.Description.str.contains(self.description_contains)]

        self.output.value = f"{len(dataf)} processes shown"

        # Add HTML links.
        dataf["PK"] = dataf["PK"].apply(
            lambda x: f"""<a href={self.path_to_root}aiidalab-widgets-base/notebooks/process.ipynb?id={x} target="_blank">{x}</a>"""
        )
        self.table.value += dataf.to_html(classes="df", escape=False, index=False)

    @tl.validate("incoming_node")
    def _validate_incoming_node(self, provided):
        """Validate incoming node."""
        node_uuid = provided["value"]
        try:
            _ = uuid.UUID(node_uuid, version=4)
        except ValueError:
            self.output.value = f"""'<span style="color:red">{node_uuid}</span>'
            is not a valid UUID."""
        else:
            return node_uuid

    @tl.validate("outgoing_node")
    def _validate_outgoing_node(self, provided):
        """Validate outgoing node. The function orm.load_node takes care of managing ids and uuids."""
        node_uuid = provided["value"]
        try:
            _ = uuid.UUID(node_uuid, version=4)
        except ValueError:
            self.output.value = f"""'<span style="color:red">{node_uuid}</span>'
            is not a valid UUID."""
        else:
            return node_uuid

    @tl.default("process_label")
    def _default_process_label(self):
        return None

    @tl.validate("process_label")
    def _validate_process_label(self, provided):
        if provided["value"]:
            return provided["value"]
        return None

    def _follow(self, update_interval):
        while True:
            self.update()
            time.sleep(update_interval)

    def start_autoupdate(self, update_interval=10):
        import threading

        update_state = threading.Thread(target=self._follow, args=(update_interval,))
        update_state.start()


class ProcessMonitor(tl.HasTraits):
    """Monitor a process and execute callback functions at specified intervals."""

    value = tl.Unicode(allow_none=True)

    def __init__(self, callbacks=None, on_sealed=None, timeout=None, **kwargs):
        self.callbacks = [] if callbacks is None else list(callbacks)
        self.on_sealed = [] if on_sealed is None else list(on_sealed)
        self.timeout = 1.0 if timeout is None else timeout

        self._monitor_thread = None
        self._monitor_thread_stop = threading.Event()
        self._monitor_thread_lock = threading.Lock()

        self.log_widget: ipw.Output | None = kwargs.pop("log_widget", None)

        super().__init__(**kwargs)

    @tl.observe("value")
    def _observe_process(self, change):
        """When the value (process uuid) is changed, stop the previous
        monitor if exist. Start a new one in thread."""
        process_uuid = change["new"]

        # stop thread (if running)
        if self._monitor_thread is not None:
            with self._monitor_thread_lock:
                self._monitor_thread_stop.set()
                self._monitor_thread.join()

        if process_uuid is None:
            return

        with self._monitor_thread_lock:
            self._monitor_thread_stop.clear()
            self._monitor_thread = threading.Thread(
                target=self._monitor_process, args=(process_uuid,)
            )
            self._monitor_thread.start()

    def _monitor_process(self, process_uuid):
        assert process_uuid is not None
        process = orm.load_node(process_uuid)

        disabled_funcs = set()

        def _run(funcs):
            for func in funcs:
                # skip all functions that had previously raised an exception
                if func in disabled_funcs:
                    continue

                try:
                    if len(inspect.signature(func).parameters) > 0:
                        func(process_uuid)
                    else:
                        func()
                except Exception:
                    if self.log_widget:
                        with self.log_widget:
                            traceback.print_exc(file=sys.stdout)
                    warnings.warn(
                        f"WARNING: The callback function {func.__name__!r} was disabled due to an error:\n{traceback.format_exc()}",
                        stacklevel=2,
                    )
                    disabled_funcs.add(func)

        while not process.is_sealed:
            _run(self.callbacks)

            if self._monitor_thread_stop.wait(timeout=self.timeout):
                break  # thread was signaled to be stopped

        # Final update:
        _run(self.callbacks)

        # Run special 'on_sealed' callback functions in case that process is sealed.
        if process.is_sealed:
            _run(self.on_sealed)

    def join(self):
        if self._monitor_thread is not None:
            self._monitor_thread.join()


class ProcessNodesTreeWidget(ipw.VBox):
    """A tree widget for the structured representation of a process graph."""

    value = tl.Unicode(allow_none=True)
    selected_nodes = tl.Tuple(read_only=True).tag(trait=tl.Instance(orm.Node))

    def __init__(self, title="Process Tree", **kwargs):
        self.title = title  # needed for ProcessFollowerWidget

        self._tree = NodesTreeWidget()
        self._tree.observe(self._observe_tree_selected_nodes, ["selected_nodes"])
        super().__init__(children=[self._tree], **kwargs)
        self.update()

    def _observe_tree_selected_nodes(self, change):
        self.set_trait("selected_nodes", change["new"])

    def update(self, _=None):
        self._tree.update()

    @tl.observe("value")
    def _observe_process(self, change):
        process_uuid = change["new"]
        if process_uuid:
            process = orm.load_node(process_uuid)
            self._tree.nodes = [process]
            self._tree.find_node(process.pk).selected = True
        else:
            self._tree.nodes = []



"""Useful functions that provide access to some data."""

import ipywidgets as ipw
import numpy as np
from ase import Atom, Atoms

# The first atom is anchoring, so the new bond will be connecting it
# The direction of the new bond is (-1, -1, -1).
FUNCTIONAL_GROUPS = {
    "Select functional group": 0,
    "Methyl -CH3": [
        ("C", 0, 0, 0),
        ("H", 0.23962342, -0.47699124, 0.78585262),
        ("H", 0.78584986, 0.23962732, -0.47698795),
        ("H", -0.47699412, 0.78585121, 0.23962671),
    ],
    "Methylene =CH2": [
        ("C", 0, 0, 0),
        ("H", -0.39755349, 0.59174911, 0.62728004),
        ("H", 0.94520686, -0.04409933, -0.07963039),
    ],
    "Hydroxy -OH": [("O", 0, 0, 0), ("H", 0.87535922, -0.3881659, 0.06790889)],
    "Amine -NH2": [
        ("N", 0, 0, 0),
        ("H", 0.7250916, -0.56270993, 0.42151063),
        ("H", -0.56261958, 0.4215284, 0.72515241),
    ],
}


class FunctionalGroupSelectorWidget(ipw.Dropdown):
    """Class to select functional groups that are returned as `Atoms` object"""

    def __init__(self, value=0, description="Select functional group", **kwargs):
        self.style = {"description_width": "initial"}
        self.layout = {"width": "initial"}
        super().__init__(
            value=value,
            description=description,
            options=tuple((key, value) for key, value in FUNCTIONAL_GROUPS.items()),
            **kwargs,
        )

    def rotate(self, align_to=(0, 0, 1), remove_anchor=False):
        """Rotate group in such a way that vector which was (-1,-1,-1)
        is alligned with align_to."""

        vect = np.array(align_to)
        norm = np.linalg.norm(vect)

        if self.value == 0:
            return None

        mol = Atoms()
        for atm in self.value:
            mol.append(Atom(atm[0], atm[1:]))

        # Bad cases.
        if norm == 0.0:
            vect = np.array((1, 1, 1)) / np.sqrt(3)
        else:
            vect = vect / norm

        mol.rotate((1, 1, 1), vect)

        if remove_anchor:
            del mol[0]

        return mol

    @property
    def anchoring_atom(self):
        """Return anchoring atom chemical symbol."""
        if self.value == 0:
            return None
        return self.value[0][0]


# LigandSelectorWidget is deprecated
LigandSelectorWidget = FunctionalGroupSelectorWidget



from __future__ import annotations

"""Jupyter viewers for AiiDA data objects."""

import base64
import copy
import os
import pathlib
import re
import warnings

import ase
import ipywidgets as ipw
import nglview
import numpy as np
import shortuuid
import spglib
import traitlets as tl
import vapory
from aiida import cmdline, orm, tools
from aiida.orm.nodes.data.structure import _get_dimensionality
from ase.data import colors
from IPython.display import clear_output, display
from matplotlib.colors import to_rgb

from .dicts import RGB_COLORS, Colors, Radius
from .loaders import LoadingWidget
from .misc import CopyToClipboardButton, ReversePolishNotation
from .utils import ase2spglib, list_to_string_range, string_range_to_list

AIIDA_VIEWER_MAPPING = {}


def register_viewer_widget(key):
    """Register widget as a viewer for the given key."""

    def registration_decorator(widget):
        AIIDA_VIEWER_MAPPING[key] = widget
        return widget

    return registration_decorator


def viewer(obj, **kwargs):
    """Display AiiDA data types in Jupyter notebooks.

    Returns the object itself if the viewer wasn't found."""
    if not isinstance(obj, orm.Node):  # only working with AiiDA nodes
        warnings.warn(
            f"This viewer works only with AiiDA objects, got {type(obj)}", stacklevel=2
        )
        return obj

    _viewer = AIIDA_VIEWER_MAPPING.get(obj.node_type)
    if isinstance(obj, orm.ProcessNode):
        # Allow to register specific viewers based on obj.process_type
        _viewer = AIIDA_VIEWER_MAPPING.get(obj.process_type, _viewer)

    if _viewer:
        return _viewer(obj, **kwargs)
    else:
        # No viewer registered for this type, return object itself
        return obj


class AiidaNodeViewWidget(ipw.VBox):
    node = tl.Instance(orm.Node, allow_none=True)

    def __init__(self, **kwargs):
        self._output = ipw.Output()
        self.node_views = {}
        self.node_view_loading_message = LoadingWidget("Loading node view")
        super().__init__(**kwargs)
        self.add_class("aiida-node-view-widget")

    @tl.observe("node")
    def _observe_node(self, change):
        if not ((node := change["new"]) and node != change["old"]):
            return
        if node.uuid in self.node_views:
            self.children = [self.node_views[node.uuid]]
            return
        self.children = [self.node_view_loading_message]
        node_view = viewer(node)
        if isinstance(node_view, ipw.DOMWidget):
            self.node_views[node.uuid] = node_view
            self.children = [node_view]
        else:
            with self._output:
                clear_output()
                if change["new"]:
                    display(node_view)
            self.children = [self._output]


@register_viewer_widget("data.core.dict.Dict.")
class DictViewer(ipw.VBox):
    value = tl.Unicode()
    """Viewer class for Dict object.

    :param parameter: Dict object to be viewed
    :type parameter: Dict
    :param downloadable: If True, add link/button to download the content of the object
    :type downloadable: bool"""

    def __init__(self, parameter, downloadable=True, **kwargs):
        import pandas as pd

        # Here we are defining properties of 'df' class (specified while exporting pandas table into html).
        # Since the exported object is nothing more than HTML table, all 'standard' HTML table settings
        # can be applied to it as well.
        # For more information on how to controle the table appearance please visit:
        # https://css-tricks.com/complete-guide-table-element/
        self.widget = ipw.HTML()
        ipw.dlink((self, "value"), (self.widget, "value"))

        self.value += """
        <style>
            .df { border: none; }
            .df tbody tr:nth-child(odd) { background-color: #e5e7e9; }
            .df tbody tr:nth-child(odd):hover { background-color:   #f5b7b1; }
            .df tbody tr:nth-child(even):hover { background-color:  #f5b7b1; }
            .df tbody td { min-width: 300px; text-align: center; border: none }
            .df th { text-align: center; border: none;  border-bottom: 1px solid black;}
        </style>
        """

        pd.set_option("max_colwidth", 40)
        dataf = pd.DataFrame(
            sorted(parameter.get_dict().items()),
            columns=["Key", "Value"],
        )
        self.value += dataf.to_html(
            classes="df", index=False
        )  # specify that exported table belongs to 'df' class
        # this is used to setup table's appearance using CSS
        if downloadable:
            payload = base64.b64encode(dataf.to_csv(index=False).encode()).decode()
            fname = f"{parameter.pk}.csv"
            self.value += f"""Download table in csv format: <a download="{fname}"
            href="data:text/csv;base64,{payload}" target="_blank">{fname}</a>"""

        super().__init__([self.widget], **kwargs)


class NglViewerRepresentation(ipw.HBox):
    """This class represents the parameters for displaying a structure in NGLViewer.

    It is utilized in the structure viewer, where multiple representations can be defined,
    each specifying how to visually represent a particular subset of atoms.
    """

    viewer_class = None  # The structure viewer class that contains this representation.

    def __init__(self, style_id, indices=None, deletable=True, atom_show_threshold=1):
        """Initialize the representation.

        style_id: str
            Unique identifier for the representation.
        indices: list
            List of indices to be displayed.
        deletable: bool
            If True, add a button to delete the representation.
        atom_show_threshold: int
            only show the atom if the corresponding value in the representation array is larger or equal than this threshold.
        """

        self.atom_show_threshold = atom_show_threshold
        self.style_id = style_id

        self.show = ipw.Checkbox(
            value=True,
            layout={"width": "40px"},
            style={"description_width": "0px"},
            disabled=False,
        )

        self.selection = ipw.Text(
            value=list_to_string_range(indices, shift=1) if indices is not None else "",
            layout={"width": "80px"},
            style={"description_width": "0px"},
        )
        self.type = ipw.Dropdown(
            options=["ball+stick", "spacefill"],
            value="ball+stick",
            disabled=False,
            layout={"width": "100px"},
            style={"description_width": "0px"},
        )
        self.size = ipw.FloatText(
            value=3,
            layout={"width": "40px"},
            style={"description_width": "0px"},
        )
        self.color = ipw.Dropdown(
            options=["element", "red", "green", "blue", "yellow", "orange", "purple"],
            value="element",
            disabled=False,
            layout={"width": "80px"},
            style={"description_width": "initial"},
        )

        # Delete button.
        self.delete_button = ipw.Button(
            description="",
            icon="trash",
            button_style="danger",
            layout={
                "width": "50px",
                "visibility": "visible" if deletable else "hidden",
            },
        )
        self.delete_button.on_click(self.delete_myself)

        super().__init__(
            children=[
                self.show,
                self.selection,
                self.type,
                self.size,
                self.color,
                self.delete_button,
            ]
        )

    def delete_myself(self, _):
        self.viewer_class.delete_representation(self)

    def sync_myself_to_array_from_atoms_object(self, structure: ase.Atoms | None):
        """Update representation from the structure object."""
        if structure:
            if self.style_id in structure.arrays:
                self.selection.value = list_to_string_range(
                    np.where(self.atoms_in_representation(structure))[0], shift=1
                )

    def add_myself_to_atoms_object(self, structure: ase.Atoms | None):
        """Add representation array to the structure object. If the array already exists, update it."""
        if structure:
            array_representation = np.full(len(structure), -1, dtype=int)
            selection = np.array(
                string_range_to_list(self.selection.value, shift=-1)[0], dtype=int
            )
            # Only attempt to display the existing atoms.
            array_representation[selection[selection < len(structure)]] = 1
            structure.set_array(self.style_id, array_representation)

    def atoms_in_representation(self, structure: ase.Atoms | None = None):
        """Return an array of booleans indicating which atoms are present in the representation."""
        if structure and self.style_id in structure.arrays:
            return structure.arrays[self.style_id] >= self.atom_show_threshold
        natoms = 0 if not structure else len(structure)
        return np.zeros(natoms, dtype=bool)

    def nglview_parameters(self, indices):
        """Return the parameters dictionary of a representation."""
        nglview_parameters_dict = {
            "type": "spacefill",
            "params": {
                "sele": "@" + ",".join(map(str, indices))
                if len(indices) > 0
                else "none",
                "opacity": 1,
                "color": self.color.value,
            },
        }
        if self.type.value == "ball+stick":
            nglview_parameters_dict["params"]["radiusScale"] = self.size.value * 0.08
        elif self.type.value == "spacefill":
            nglview_parameters_dict["params"]["radiusScale"] = self.size.value * 0.25

        return nglview_parameters_dict


class _StructureDataBaseViewer(ipw.VBox):
    """Base viewer class for AiiDA structure or trajectory objects.

    Traits:
        _all_representations: list, containing all the representations of the structure.
        input_selection: list used mostly by external tools to populate the selection field.
        selection: list of currently selected atoms.
        displayed_selection: list of currently displayed atoms in the displayed structure, which also includes super-cell.
        supercell: list of supercell dimensions.
        cell: ase.cell.Cell object.
    """

    _all_representations = tl.List()
    input_selection = tl.List(tl.Int(), allow_none=True)
    selection = tl.List(tl.Int())
    displayed_selection = tl.List(tl.Int())
    supercell = tl.List(tl.Int())
    cell = tl.Instance(ase.cell.Cell, allow_none=True)
    DEFAULT_SELECTION_OPACITY = 0.2
    DEFAULT_SELECTION_RADIUS = 6
    DEFAULT_SELECTION_COLOR = "green"
    REPRESENTATION_PREFIX = "_aiidalab_viewer_representation_"
    DEFAULT_REPRESENTATION = "_aiidalab_viewer_representation_default"
    _CELL_LABELS = {
        1: ["length", "Å"],
        2: ["area", "Å²"],
        3: ["volume", "Å³"],
    }

    def __init__(
        self,
        configure_view=True,
        configuration_tabs=None,
        default_camera="orthographic",
        **kwargs,
    ):
        """Initialize the viewer.

        :param configure_view: If True, add configuration tabs (deprecated).
        :param configuration_tabs: List of configuration tabs (default: ["Selection", "Appearance", "Cell", "Download"]).
        :param default_camera: default camera (orthographic|perspective), can be changed in the Appearance tab.
        """
        # Defining viewer box.

        # Nglviwer
        self._viewer = nglview.NGLWidget()
        self._viewer.camera = default_camera
        self._viewer.observe(self._on_atom_click, names="picked")
        self._viewer.stage.set_parameters(mouse_preset="pymol")

        view_box = ipw.VBox([self._viewer])
        view_box.add_class("view-box")

        configuration_tabs_map = {
            "Cell": self._cell_tab(),
            "Selection": self._selection_tab(),
            "Appearance": self._appearance_tab(),
            "Download": self._download_tab(),
        }

        if configure_view is not True:
            warnings.warn(
                "`configure_view` is deprecated, please use `configuration_tabs` instead.",
                DeprecationWarning,
                stacklevel=2,
            )
            if not configure_view:
                configuration_tabs.clear()

        # Constructing configuration box
        if configuration_tabs is None:
            configuration_tabs = ["Selection", "Appearance", "Cell", "Download"]
        if len(configuration_tabs) != 0:
            self.configuration_box = ipw.Tab(
                layout=ipw.Layout(flex="1 1 auto", width="auto")
            )
            self.configuration_box.children = [
                configuration_tabs_map[tab_title] for tab_title in configuration_tabs
            ]

            for i, title in enumerate(configuration_tabs):
                self.configuration_box.set_title(i, title)
            children = [ipw.HBox([view_box, self.configuration_box])]
            view_box.layout = {"width": "60%"}
        else:
            children = [view_box]

        if "children" in kwargs:
            children += kwargs.pop("children")

        super().__init__(children, **kwargs)

    def _selection_tab(self):
        """Defining the selection tab."""

        # 1. Selected atoms.
        self._selected_atoms = ipw.Text(
            description="Select atoms:",
            value="",
            style={"description_width": "initial"},
        )

        # 2. Copy to clipboard
        copy_to_clipboard = CopyToClipboardButton(description="Copy to clipboard")
        tl.link((self._selected_atoms, "value"), (copy_to_clipboard, "value"))

        # 3. Informing about wrong syntax.
        self.wrong_syntax = ipw.HTML(
            value="""<i class="fa fa-times" style="color:red;font-size:2em;" ></i> wrong syntax""",
            layout={"visibility": "hidden"},
        )

        # 4. Button to clear selection.
        clear_selection = ipw.Button(description="Clear selection")
        clear_selection.on_click(
            lambda _: self.set_trait("displayed_selection", [])
        )  # lambda cannot contain assignments

        # 5. Button to apply selection
        apply_displayed_selection = ipw.Button(description="Apply selection")
        apply_displayed_selection.on_click(self.apply_displayed_selection)

        self.selection_info = ipw.HTML()

        return ipw.VBox(
            [
                ipw.HBox([self._selected_atoms, self.wrong_syntax]),
                ipw.HTML(
                    value="""
                <p style="font-weight:800;">You can either specify ranges:
                    <font style="font-style:italic;font-weight:400;">1 5..8 10</font>
                </p>
                <p style="font-weight:800;">or expressions:
                    <font style="font-style:italic;font-weight:400;">(x>1 and name not [N,O]) or d_from [1,1,1]>2 or id>=10</font>
                </p>"""
                ),
                ipw.HBox(
                    [copy_to_clipboard, clear_selection, apply_displayed_selection]
                ),
                self.selection_info,
            ]
        )

    def _appearance_tab(self):
        """Defining the appearance tab."""

        # 1. Supercell
        def change_supercell(_=None):
            self.supercell = [
                _supercell[0].value,
                _supercell[1].value,
                _supercell[2].value,
            ]

        _supercell = [
            ipw.BoundedIntText(value=1, min=1, layout={"width": "40px"}),
            ipw.BoundedIntText(value=1, min=1, layout={"width": "40px"}),
            ipw.BoundedIntText(value=1, min=1, layout={"width": "40px"}),
        ]
        for elem in _supercell:
            elem.observe(change_supercell, names="value")
        supercell_selector = ipw.HBox(
            [
                ipw.HTML(
                    description="Supercell:", style={"description_width": "initial"}
                ),
                *_supercell,
            ]
        )

        # 2. Choose background color.
        background_color = ipw.ColorPicker(
            description="Background",
            style={"description_width": "initial"},
            layout={"width": "200px"},
        )
        tl.link((background_color, "value"), (self._viewer, "background"))
        background_color.value = "white"

        # 3. Camera switcher
        camera_type = ipw.ToggleButtons(
            options=[("Orthographic", "orthographic"), ("Perspective", "perspective")],
            description="Camera type:",
            value=self._viewer.camera,
            layout={"align_self": "flex-start"},
            style={"button_width": "115.5px"},
        )

        def change_camera(change):
            self._viewer.camera = change["new"]

        camera_type.observe(change_camera, names="value")

        # 4. Center button.
        center_button = ipw.Button(description="Center molecule")
        center_button.on_click(lambda _: self._viewer.center())

        # 5. representations buttons
        self.representations_header = ipw.HBox(
            [
                ipw.HTML(
                    """<p style="text-align:center">Show</p>""",
                    layout={"width": "40px"},
                ),
                ipw.HTML(
                    """<p style="text-align:center">Atoms</p>""",
                    layout={"width": "80px"},
                ),
                ipw.HTML(
                    """<p style="text-align:center">Type</p>""",
                    layout={"width": "100px"},
                ),
                ipw.HTML(
                    """<p style="text-align:center">Size</p>""",
                    layout={"width": "40px"},
                ),
                ipw.HTML(
                    """<p style="text-align:center">Color</p>""",
                    layout={"width": "80px"},
                ),
                ipw.HTML(
                    """<p style="text-align:center">Delete</p>""",
                    layout={"width": "50px"},
                ),
            ]
        )
        self.atoms_not_represented = ipw.HTML()
        add_new_representation_button = ipw.Button(
            description="Add representation", button_style="info"
        )
        add_new_representation_button.on_click(self._add_representation)

        apply_representations = ipw.Button(description="Apply representations")
        apply_representations.on_click(self._apply_representations)
        self.representation_output = ipw.VBox()

        # The default representation is always present and cannot be deleted.
        self._all_representations = [
            NglViewerRepresentation(
                style_id=self.DEFAULT_REPRESENTATION,
                deletable=False,
                atom_show_threshold=0,
            )
        ]

        representation_accordion = ipw.Accordion(
            children=[
                ipw.VBox(
                    [
                        self.representations_header,
                        self.representation_output,
                        self.atoms_not_represented,
                        ipw.HBox(
                            [apply_representations, add_new_representation_button]
                        ),
                    ]
                )
            ],
        )
        representation_accordion.set_title(0, "Representations")
        representation_accordion.selected_index = None

        info = ipw.HTML("""
            <div style="line-height: 1.5; max-width: 460px; margin: 6px 2px;">
                <b>Note:</b> The supercell settings here are <b>for visualization
                purposes only</b>. To simulate a supercell, open the <b>Edit
                structure</b> panel below, navigate to the <b>Edit cell</b> tab, and
                use the <b>Cell transformation</b> controls.
            </div>
        """)

        return ipw.VBox(
            [
                supercell_selector,
                info,
                background_color,
                camera_type,
                center_button,
                representation_accordion,
            ]
        )

    def _add_representation(self, _=None, style_id=None, indices=None):
        """Add a representation to the list of representations."""
        self._all_representations = [
            *self._all_representations,
            NglViewerRepresentation(
                style_id=style_id or f"{self.REPRESENTATION_PREFIX}{shortuuid.uuid()}",
                indices=indices,
            ),
        ]
        self._apply_representations()

    def delete_representation(self, representation: NglViewerRepresentation):
        try:
            index = self._all_representations.index(representation)
        except ValueError:
            self.representation_add_message.message = f"""<span style="color:red">Error:</span> Rep. {representation} not found."""
            return

        self._all_representations = (
            self._all_representations[:index] + self._all_representations[index + 1 :]
        )

        if representation.style_id in self.structure.arrays:
            del self.structure.arrays[representation.style_id]
        del representation
        self._apply_representations()

    @tl.observe("_all_representations")
    def _observe_all_representations(self, change):
        """Update the list of representations."""
        self.representation_output.children = change["new"]
        if change["new"]:
            self._all_representations[-1].viewer_class = self

    def _povray_cylinder(self, v1, v2, radius, color):
        """Create a cylinder for POVRAY."""
        return vapory.Cylinder(
            v1,
            v2,
            radius,
            vapory.Pigment("color", color),
            vapory.Finish("phong", 0.8, "reflection", 0.05),
        )

    def _cylinder(self, v1, v2, radius, color):
        """Create a cylinder for NGLViewer."""
        return (
            "cylinder",
            tuple(v1),
            tuple(v2),
            tuple(color),
            radius,
        )

    def _compute_bonds(self, structure, radius=1.0, color="element", povray=False):
        """Create an list of bonds for the structure."""

        import ase.neighborlist

        bonds = []
        if len(structure) <= 1:
            return []
        # The radius is scaled by 0.04 to have a better visual appearance.
        radius = radius * 0.04

        # The value 1.09 is chosen based on our experience. It is a good compromise between showing too many bonds
        # and not showing bonds that should be there.
        cutoff = ase.neighborlist.natural_cutoffs(structure, mult=1.09)

        ii, bond_vectors = ase.neighborlist.neighbor_list(
            "iD", structure, cutoff, self_interaction=False
        )
        nb = len(ii)
        # bond start position
        v1 = structure.positions[ii]
        # middle position
        v2 = v1 + bond_vectors * 0.5

        # Choose the correct way for computing the cylinder.
        if povray:
            symbols = structure.get_chemical_symbols()
            bonds = [
                self._povray_cylinder(v1[ib], v2[ib], radius, Colors[symbols[ii[ib]]])
                for ib in range(nb)
            ]
        else:
            if color == "element":
                numbers = structure.get_atomic_numbers()
                bonds = [
                    self._cylinder(
                        v1[ib], v2[ib], radius, colors.jmol_colors[numbers[ii[ib]]]
                    )
                    for ib in range(nb)
                ]
            else:
                bonds = [
                    self._cylinder(v1[ib], v2[ib], radius, RGB_COLORS[color])
                    for ib in range(nb)
                ]
        return bonds

    def _apply_representations(self, change=None):
        """Apply the representations to the displayed structure."""
        representation_ids = []

        # Representation can only be applied if a structure is present.
        if self.structure is None:
            return

        # Add existing representations to the structure.
        for representation in self._all_representations:
            representation.add_myself_to_atoms_object(self.structure)
            representation_ids.append(representation.style_id)

        # Remove missing representations from the structure.
        for array in self.structure.arrays:
            if (
                array.startswith(self.REPRESENTATION_PREFIX)
                and array not in representation_ids
            ):
                del self.structure.arrays[array]
        self._observe_structure({"new": self.structure})
        self._check_missing_atoms_in_representations()

    def _check_missing_atoms_in_representations(self):
        missing_atoms = np.zeros(self.natoms)
        for rep in self._all_representations:
            missing_atoms += rep.atoms_in_representation(self.structure)
        missing_atoms = np.where(missing_atoms == 0)[0]
        if len(missing_atoms) > 0:
            self.atoms_not_represented.value = (
                "Atoms excluded from representations: "
                + list_to_string_range(list(missing_atoms), shift=1)
            )
        else:
            self.atoms_not_represented.value = ""

    @tl.observe("cell")
    def _observe_cell(self, _=None):
        # Updtate the Cell and Periodicity.
        if self.cell:
            self.cell_a.value = "<i><b>a</b></i>: {:.4f} {:.4f} {:.4f}".format(
                *self.cell.array[0]
            )
            self.cell_b.value = "<i><b>b</b></i>: {:.4f} {:.4f} {:.4f}".format(
                *self.cell.array[1]
            )
            self.cell_c.value = "<i><b>c</b></i>: {:.4f} {:.4f} {:.4f}".format(
                *self.cell.array[2]
            )

            self.cell_a_length.value = (
                f"|<i><b>a</b></i>|: {self.cell.lengths()[0]:.4f}"
            )
            self.cell_b_length.value = (
                f"|<i><b>b</b></i>|: {self.cell.lengths()[1]:.4f}"
            )
            self.cell_c_length.value = (
                f"|<i><b>c</b></i>|: {self.cell.lengths()[2]:.4f}"
            )

            self.cell_alpha.value = f"&alpha;: {self.cell.angles()[0]:.4f}"
            self.cell_beta.value = f"&beta;: {self.cell.angles()[1]:.4f}"
            self.cell_gamma.value = f"&gamma;: {self.cell.angles()[2]:.4f}"

            spglib_structure = ase2spglib(self.structure)
            symmetry_dataset = spglib.get_symmetry_dataset(
                spglib_structure, symprec=1e-5, angle_tolerance=1.0
            )

            periodicity_map = {
                (True, True, True): "xyz",
                (True, False, False): "x",
                (False, True, False): "y",
                (False, False, True): "z",
                (True, True, False): "xy",
                (True, False, True): "xz",
                (False, True, True): "yz",
                (False, False, False): "-",
            }
            self.cell_spacegroup.value = f"Spacegroup: {symmetry_dataset['international']} (No.{symmetry_dataset['number']})"
            self.cell_hall.value = f"Hall: {symmetry_dataset['hall']} (No.{symmetry_dataset['hall_number']})"
            self.periodicity.value = (
                f"Periodicity: {periodicity_map[tuple(self.structure.pbc)]}"
            )
            # Calculate the volume of the cell using the function from orm.StructureData
            dimension_data = _get_dimensionality(self.structure.pbc, self.cell)
            # Determine the label and unit based on dimensionality
            cell_label = self._CELL_LABELS.get(dimension_data["dim"])
            if cell_label:
                self.cell_volume.value = f"Cell {cell_label[0]}: {dimension_data['value']:.4f} ({cell_label[1]})"
            else:
                self.cell_volume.value = "Cell volume: -"

        else:
            self.cell_a.value = "<i><b>a</b></i>:"
            self.cell_b.value = "<i><b>b</b></i>:"
            self.cell_c.value = "<i><b>c</b></i>:"

            self.cell_a_length.value = "|<i><b>a</b></i>|:"
            self.cell_b_length.value = "|<i><b>b</b></i>|:"
            self.cell_c_length.value = "|<i><b>c</b></i>|:"

            self.cell_alpha.value = "&alpha;:"
            self.cell_beta.value = "&beta;:"
            self.cell_gamma.value = "&gamma;:"

            self.cell_spacegroup.value = ""
            self.cell_hall.value = ""
            self.periodicity.value = ""

    def _cell_tab(self):
        self.cell_a = ipw.HTML()
        self.cell_b = ipw.HTML()
        self.cell_c = ipw.HTML()

        self.cell_a_length = ipw.HTML()
        self.cell_b_length = ipw.HTML()
        self.cell_c_length = ipw.HTML()

        self.cell_alpha = ipw.HTML()
        self.cell_beta = ipw.HTML()
        self.cell_gamma = ipw.HTML()

        self.cell_spacegroup = ipw.HTML()
        self.cell_hall = ipw.HTML()
        self.periodicity = ipw.HTML()

        self.cell_volume = ipw.HTML()

        self._observe_cell()

        return ipw.VBox(
            [
                ipw.HTML("Length unit: angstrom (Å)"),
                ipw.HBox(
                    [
                        ipw.VBox(
                            [
                                ipw.HTML("Cell vectors:"),
                                self.cell_a,
                                self.cell_b,
                                self.cell_c,
                            ]
                        ),
                        ipw.VBox(
                            [
                                ipw.HTML("Сell vectors length:"),
                                self.cell_a_length,
                                self.cell_b_length,
                                self.cell_c_length,
                            ],
                            layout={"margin": "0 0 0 50px"},
                        ),
                    ]
                ),
                ipw.HBox(
                    [
                        ipw.VBox(
                            [
                                ipw.HTML("Angles:"),
                                self.cell_alpha,
                                self.cell_beta,
                                self.cell_gamma,
                            ]
                        ),
                        ipw.VBox(
                            [
                                ipw.HTML("Symmetry information:"),
                                self.cell_spacegroup,
                                self.cell_hall,
                                self.periodicity,
                            ],
                            layout={"margin": "0 0 0 50px"},
                        ),
                    ]
                ),
                self.cell_volume,
            ]
        )

    def _download_tab(self):
        """Defining the download tab."""

        # 1. Choose download file format.
        self.file_format = ipw.Dropdown(
            label="Extended xyz",
            # File extension and format may be different. Therefore, we define both.
            options=(
                ("xyz", {"extension": "xyz", "format": "xyz"}),
                ("cif", {"extension": "cif", "format": "cif"}),
                ("Extended xyz", {"extension": "xyz", "format": "extxyz"}),
                ("xsf", {"extension": "xsf", "format": "xsf"}),
            ),
            layout={"width": "200px"},
            description="File format:",
        )

        # 2. Download button.
        self.download_btn = ipw.Button(description="Download")
        self.download_btn.on_click(self.download)
        self.download_box = ipw.VBox(
            children=[
                ipw.Label("Download as file:"),
                ipw.HBox([self.file_format, self.download_btn]),
            ]
        )

        # 3. Screenshot button
        self.screenshot_btn = ipw.Button(description="Screenshot", icon="camera")
        self.screenshot_btn.on_click(lambda _: self._viewer.download_image())
        self.screenshot_box = ipw.VBox(
            children=[ipw.Label("Create a screenshot:"), self.screenshot_btn]
        )

        # 4. Render a high quality image
        self.render_btn = ipw.Button(description="Render", icon="paint-brush")
        self.render_btn.on_click(self._render_structure)
        self.render_box = ipw.VBox(
            children=[ipw.Label("Render an image with POVRAY:"), self.render_btn]
        )

        return ipw.VBox([self.download_box, self.screenshot_box, self.render_box])

    def _render_structure(self, change=None):
        """Render the structure with POVRAY."""
        from tempfile import TemporaryDirectory

        if not isinstance(self.displayed_structure, ase.Atoms):
            return

        self.render_btn.disabled = True
        omat = np.array(self._viewer._camera_orientation).reshape(4, 4).transpose()

        zfactor = np.linalg.norm(omat[0, 0:3])
        omat[0:3, 0:3] = omat[0:3, 0:3] / zfactor

        bb = copy.deepcopy(self.displayed_structure)
        bb.pbc = (False, False, False)

        for i in bb:
            ixyz = omat[0:3, 0:3].dot(np.array([i.x, i.y, i.z]) + omat[0:3, 3])
            i.x, i.y, i.z = -ixyz[0], ixyz[1], ixyz[2]

        vertices = []

        cell = bb.get_cell()
        vertices.append(np.array([0, 0, 0]))
        vertices.extend(cell)
        vertices.extend(
            [
                cell[0] + cell[1],
                cell[0] + cell[2],
                cell[1] + cell[2],
                cell[0] + cell[1] + cell[2],
            ]
        )

        for n, i in enumerate(vertices):
            ixyz = omat[0:3, 0:3].dot(i + omat[0:3, 3])
            vertices[n] = np.array([-ixyz[0], ixyz[1], ixyz[2]])

        bonds = self._compute_bonds(bb, povray=True)

        edges = []
        for x, i in enumerate(vertices):
            for j in vertices[x + 1 :]:
                if (
                    np.linalg.norm(np.cross(i - j, vertices[1] - vertices[0])) < 0.001
                    or np.linalg.norm(np.cross(i - j, vertices[2] - vertices[0]))
                    < 0.001
                    or np.linalg.norm(np.cross(i - j, vertices[3] - vertices[0]))
                    < 0.001
                ):
                    edge = vapory.Cylinder(
                        i,
                        j,
                        0.06,
                        vapory.Texture(
                            vapory.Pigment(
                                "color", [212 / 255.0, 175 / 255.0, 55 / 255.0]
                            )
                        ),
                        vapory.Finish("phong", 0.9, "reflection", 0.01),
                    )
                    edges.append(edge)

        camera = vapory.Camera(
            "perspective",
            "location",
            [0, 0, -zfactor / 1.5],
            "look_at",
            [0.0, 0.0, 0.0],
        )
        light = vapory.LightSource([0, 0, -100.0], "color", [1.5, 1.5, 1.5])

        spheres = [
            vapory.Sphere(
                [i.x, i.y, i.z],
                Radius[i.symbol],
                vapory.Texture(vapory.Pigment("color", np.array(Colors[i.symbol]))),
                vapory.Finish("phong", 0.9, "reflection", 0.05),
            )
            for i in bb
        ]

        objects = [
            light,
            *spheres,
            *edges,
            *bonds,
            vapory.Background("color", np.array(to_rgb(self._viewer.background))),
        ]

        scene = vapory.Scene(camera, objects=objects)
        fname = bb.get_chemical_formula() + ".png"

        with TemporaryDirectory() as tmpdir:
            cwd = pathlib.Path.cwd()
            try:
                os.chdir(tmpdir)
                scene.render(
                    fname,
                    width=2560,
                    height=1440,
                    antialiasing=0.000,
                    quality=11,
                    remove_temp=False,
                )
                with open(fname, "rb") as raw:
                    payload = base64.b64encode(raw.read()).decode()
            finally:
                os.chdir(cwd)

        self._download(payload=payload, filename=fname)
        self.render_btn.disabled = False

    def _on_atom_click(self, _=None):
        """Update selection when clicked on atom."""
        if hasattr(self._viewer, "component_0"):
            # Did not click on atom:
            if "atom1" not in self._viewer.picked.keys():
                return

            index = self._viewer.picked["atom1"]["index"]

            displayed_selection = self.displayed_selection.copy()
            if displayed_selection:
                if index not in displayed_selection:
                    displayed_selection.append(index)
                else:
                    displayed_selection.remove(index)
            else:
                displayed_selection = [index]
            self.displayed_selection = displayed_selection

    def highlight_atoms(
        self,
        list_of_atoms,
    ):
        """Highlighting atoms according to the provided list."""
        if not hasattr(self._viewer, "component_0"):
            return

        # Create the dictionaries for highlight_representations.
        for i, representation in enumerate(self._all_representations):
            # First remove the previous highlight_representation.
            self._viewer._remove_representations_by_name(
                repr_name=f"highlight_representation_{i}", component=0
            )

            # Then add the new one if needed.
            indices = np.intersect1d(
                list_of_atoms,
                np.where(
                    representation.atoms_in_representation(self.displayed_structure)
                )[0],
            )
            if len(indices) > 0:
                params = representation.nglview_parameters(indices)
                params["params"]["name"] = f"highlight_representation_{i}"
                params["params"]["opacity"] = 0.8
                params["params"]["color"] = "darkgreen"
                params["params"]["component_index"] = 0

                # Use directly the remote call for more flexibility.
                self._viewer._remote_call(
                    "addRepresentation",
                    target="compList",
                    args=[params["type"]],
                    kwargs=params["params"],
                )

    def remove_viewer_components(self, c=None):
        """Remove all components from the viewer except the one specified."""
        if hasattr(self._viewer, "component_0"):
            self._viewer.component_0.clear_representations()
            cid = self._viewer.component_0.id
            self._viewer.remove_component(cid)

    @tl.default("supercell")
    def _default_supercell(self):
        return [1, 1, 1]

    @tl.observe("input_selection")
    def _observe_input_selection(self, value):
        if value["new"] is None:
            return

        # Exclude everything that is beyond the total number of atoms.
        selection_list = [x for x in value["new"] if x < self.natoms]

        # In the case of a supercell, we need to multiply the selection as well
        multiplier = sum(self.supercell) - 2
        selection_list = [
            x + self.natoms * i for x in selection_list for i in range(multiplier)
        ]

        self.displayed_selection = selection_list

    @tl.observe("displayed_selection")
    def _observe_displayed_selection(self, _=None):
        seen = set()
        seq = [x % self.natoms for x in self.displayed_selection]
        self.selection = [x for x in seq if not (x in seen or seen.add(x))]
        self.highlight_atoms(self.displayed_selection)

    def apply_displayed_selection(self, _=None):
        """Apply selection specified in the text field."""
        expanded_selection, syntax_ok = string_range_to_list(
            self._selected_atoms.value, shift=-1
        )
        if not syntax_ok:
            try:
                sel = self._parse_advanced_selection(
                    condition=self._selected_atoms.value
                )
                sel = list_to_string_range(sel, shift=1)
                expanded_selection, syntax_ok = string_range_to_list(sel, shift=-1)
            except (IndexError, TypeError, AttributeError):
                syntax_ok = False
                self.wrong_syntax.layout.visibility = "visible"

        if syntax_ok:
            self.wrong_syntax.layout.visibility = "hidden"
            self.displayed_selection = expanded_selection
        else:
            self.wrong_syntax.layout.visibility = "visible"

    def download(self, _=None):
        """Prepare a structure for downloading."""
        payload = self._prepare_payload(self.file_format.value["format"])
        if payload is None:
            return
        suffix = f"pk-{self.pk}" if self.pk else "not-stored"
        self._download(
            payload=payload,
            filename=f"""structure-{suffix}.{self.file_format.value["extension"]}""",
        )

    @staticmethod
    def _download(payload, filename):
        """Download payload as a file named as filename."""
        from IPython.display import Javascript

        javas = Javascript(
            f"""
            var link = document.createElement('a');
            link.href = "data:;base64,{payload}"
            link.download = "{filename}"
            document.body.appendChild(link);
            link.click();
            document.body.removeChild(link);
            """
        )
        display(javas)

    def _prepare_payload(self, file_format=None):
        """Prepare binary information."""
        from tempfile import NamedTemporaryFile

        if not self.structure:
            return None

        file_format = file_format if file_format else self.file_format.value["format"]
        tmp = NamedTemporaryFile()
        self.structure.write(tmp.name, format=file_format)
        with open(tmp.name, "rb") as raw:
            return base64.b64encode(raw.read()).decode()

    @property
    def thumbnail(self):
        return self._prepare_payload(file_format="png")

    @property
    def natoms(self):
        """Number of atoms in the structure."""
        return 0 if self.structure is None else len(self.structure)


@register_viewer_widget("data.core.cif.CifData.")
@register_viewer_widget("data.core.structure.StructureData.")
class StructureDataViewer(_StructureDataBaseViewer):
    """Viewer class for AiiDA structure objects.

    Attributes:
        structure (Atoms, StructureData, CifData): Trait that contains a structure object,
        which was initially provided to the viewer. It can be either directly set to an
        ASE Atoms object or to AiiDA structure object containing `get_ase()` method.

        displayed_structure (Atoms): Trait that contains a structure object that is
        currently displayed (supercell, for example). The trait is generated automatically
        and can't be set outside of the class.
    """

    structure = tl.Union(
        [
            tl.Instance(ase.Atoms),
            tl.Instance(orm.StructureData),
            tl.Instance(orm.CifData),
        ],
        allow_none=True,
    )
    displayed_structure = tl.Instance(ase.Atoms, allow_none=True, read_only=True)
    pk = tl.Int(allow_none=True)

    def __init__(self, structure=None, **kwargs):
        super().__init__(**kwargs)
        self.add_class("structure-viewer")
        self.structure = structure

    @tl.observe("supercell")
    def _observe_supercell(self, _=None):
        if self.structure is not None:
            self.set_trait(
                "displayed_structure", None
            )  # To make sure the structure is always updated.
            # nglview displays structures by first saving them to a temporary "pdb" file, which necessitates
            # converting the unit cell and atomic positions into a standard form where the a-axis aligns along the x-axis.
            # This transformation can cause discrepancies between the atom positions and custom bonds calculated from the original structure.
            # To mitigate this, we transform the "displayed_structure" into the standard form prior to rendering in nglview.
            # This ensures that nglview's internal handling does not further modify the structure unexpectedly.
            standard_structure = self.structure.copy()
            standard_structure.set_cell(
                self.structure.cell.standard_form()[0], scale_atoms=True
            )
            self.set_trait(
                "displayed_structure", standard_structure.repeat(self.supercell)
            )

    @tl.validate("structure")
    def _valid_structure(self, change):
        """Update structure."""
        structure = change["value"]
        if isinstance(structure, ase.Atoms):
            self.pk = None
        elif isinstance(structure, (orm.StructureData, orm.CifData)):
            self.pk = structure.pk
            structure = structure.get_ase()

        # Add default representation array if it is not present.
        # This will make sure that the new structure is displayed at the beginning.
        if self.DEFAULT_REPRESENTATION not in structure.arrays:
            structure.set_array(
                self.DEFAULT_REPRESENTATION,
                np.zeros(len(structure), dtype=int),
            )
        return structure  # This also includes the case when the structure is None.

    @tl.observe("structure")
    def _observe_structure(self, change=None):
        """Update displayed_structure trait after the structure trait has been modified."""
        structure = change["new"]

        self._viewer.clear_representations(component=0)

        if not structure:
            self.set_trait("displayed_structure", None)
            self.set_trait("cell", None)
            return

        # Make sure that the representation arrays from structure are present in the viewer.
        structure_ids = [
            style_id
            for style_id in structure.arrays
            if style_id.startswith(self.REPRESENTATION_PREFIX)
        ]
        representation_ids = [rep.style_id for rep in self._all_representations]

        for style_id in structure_ids:
            try:
                index = representation_ids.index(style_id)
                self._all_representations[index].sync_myself_to_array_from_atoms_object(
                    structure
                )
            except ValueError:
                self._add_representation(
                    style_id=style_id,
                    indices=np.where(structure.arrays[self.style_id] >= 1)[0],
                )
        # Empty atoms selection for the representations that are not present in the structure.
        # Typically this happens when a new structure is imported.
        for i, style_id in enumerate(representation_ids):
            if style_id not in structure_ids:
                self._all_representations[i].selection.value = ""

        self._observe_supercell()  # To trigger an update of the displayed structure
        self.set_trait("cell", structure.cell)

    @tl.observe("displayed_structure")
    def _observe_displayed_structure(self, change):
        """Update the view if displayed_structure trait was modified."""
        with self.hold_trait_notifications():
            self.remove_viewer_components()
            if change["new"]:
                self._viewer.add_component(
                    nglview.ASEStructure(self.displayed_structure),
                    default_representation=False,
                    name="Structure",
                )
                bonds = []
                nglview_params = []
                for representation in self._all_representations:
                    if representation.show.value:
                        indices = np.where(
                            representation.atoms_in_representation(
                                self.displayed_structure
                            )
                        )[0]
                        nglview_params.append(
                            representation.nglview_parameters(indices)
                        )

                        # Add bonds if ball+stick representation is used.
                        if representation.type.value == "ball+stick":
                            bonds += self._compute_bonds(
                                self.displayed_structure[indices],
                                representation.size.value,
                                representation.color.value,
                            )
                self._viewer.set_representations(nglview_params, component=0)
                self._viewer.add_unitcell()
                self._viewer._add_shape(set(bonds), name="bonds")
                self._viewer.center()
                # In case of a structure with only one atom, the `center()` method will show a black sphere.
                if len(self.displayed_structure) == 1:
                    # get center of mass of the displayed structure
                    com = self.displayed_structure.get_center_of_mass()
                    # The default camera should be in the z direction, so we
                    # shift the center of the control in z direction by 1
                    com[2] -= 1
                    self._viewer.control.center(com)

        self.displayed_selection = []

    def d_from(self, operand):
        point = np.array([float(i) for i in operand[1:-1].split(",")])
        return np.linalg.norm(self.displayed_structure.positions - point, axis=1)

    def name_operator(self, operand):
        """Defining the name operator which will handle atom kind names."""
        if operand.startswith("[") and operand.endswith("]"):
            names = operand[1:-1].split(",")
        elif not operand.endswith("[") and not operand.startswith("]"):
            names = [operand]
        symbols = self.displayed_structure.get_chemical_symbols()
        return np.array([i for i, val in enumerate(symbols) if val in names])

    def not_operator(self, operand):
        """Reverting the selected atoms."""
        if operand.startswith("[") and operand.endswith("]"):
            names = operand[1:-1].split(",")
        elif not operand.endswith("[") and not operand.startswith("]"):
            names = [operand]
        return (
            "["
            + ",".join(
                list(set(self.displayed_structure.get_chemical_symbols()) - set(names))
            )
            + "]"
        )

    def _parse_advanced_selection(self, condition=None):
        """Apply advanced selection specified in the text field."""

        def addition(opa, opb):
            return opa + opb

        def subtraction(opa, opb):
            return opa - opb

        def mult(opa, opb):
            return opa * opb

        def division(opa, opb):
            if isinstance(opb, type(np.array([]))):
                if any(np.abs(opb) < 0.0001):
                    return np.array([])
            elif np.abs(opb) < 0.0001:
                return np.array([])
            return opa / opb

        def power(opa, opb):
            return opa**opb

        def greater(opa, opb):
            return np.where(opa > opb)[0]

        def lower(opa, opb):
            return np.where(opa < opb)[0]

        def equal(opa, opb):
            return np.where(opa == opb)[0]

        def notequal(opa, opb):
            return np.where(opa != opb)[0]

        def greatereq(opa, opb):
            return np.where(opa >= opb)[0]

        def lowereq(opa, opb):
            return np.where(opa <= opb)[0]

        def intersec(opa, opb):
            return np.intersect1d(opa, opb)

        def union(opa, opb):
            return np.union1d(opa, opb)

        operandsdict = {
            "x": self.displayed_structure.positions[:, 0],
            "y": self.displayed_structure.positions[:, 1],
            "z": self.displayed_structure.positions[:, 2],
            "id": np.array([atom.index + 1 for atom in self.displayed_structure]),
        }

        operatorsdict = {
            ">": {
                "function": greater,
                "priority": 0,
                "nargs": 2,
            },
            "<": {
                "function": lower,
                "priority": 0,
                "nargs": 2,
            },
            ">=": {
                "function": greatereq,
                "priority": 0,
                "nargs": 2,
            },
            "<=": {
                "function": lowereq,
                "priority": 0,
                "nargs": 2,
            },
            "and": {
                "function": intersec,
                "priority": -1,
                "nargs": 2,
            },
            "or": {
                "function": union,
                "priority": -2,
                "nargs": 2,
            },
            "+": {
                "function": addition,
                "priority": 1,
                "nargs": 2,
            },
            "-": {
                "function": subtraction,
                "priority": 1,
                "nargs": 2,
            },
            "*": {
                "function": mult,
                "priority": 2,
                "nargs": 2,
            },
            "/": {
                "function": division,
                "priority": 2,
                "nargs": 2,
            },
            "^": {
                "function": power,
                "priority": 3,
                "nargs": 2,
            },
            "==": {
                "function": equal,
                "priority": 0,
                "nargs": 2,
            },
            "!=": {
                "function": notequal,
                "priority": 0,
                "nargs": 2,
            },
            "d_from": {
                "function": self.d_from,
                "priority": 11,
                "nargs": 1,
            },  # At the moment the priority is not used.
            "name": {
                "function": self.name_operator,
                "priority": 9,
                "nargs": 1,
            },  # When changed, this should be re-assesed.
            "not": {
                "function": self.not_operator,
                "priority": 10,
                "nargs": 1,
            },
        }

        rpn = ReversePolishNotation(
            operators=operatorsdict, additional_operands=operandsdict
        )
        return list(rpn.execute(expression=condition))

    def create_selection_info(self):
        """Create information to be displayed with selected atoms"""

        if not self.displayed_selection:
            return ""

        def print_pos(pos):
            return " ".join([str(i) for i in pos.round(2)])

        def add_info(indx, atom):
            return f"<p>Id: {indx + 1}; Symbol: {atom.symbol}; Coordinates: ({print_pos(atom.position)})</p>"

        # Find geometric center.
        geom_center = print_pos(
            np.average(
                self.displayed_structure[self.displayed_selection].get_positions(),
                axis=0,
            )
        )

        info = ""

        # Report coordinates.
        if len(self.displayed_selection) == 1:
            info += add_info(
                self.displayed_selection[0],
                self.displayed_structure[self.displayed_selection[0]],
            )

        # Report coordinates, distance and center.
        elif len(self.displayed_selection) == 2:
            dist = self.displayed_structure.get_distance(*self.displayed_selection)
            distv = self.displayed_structure.get_distance(
                *self.displayed_selection, vector=True
            )
            info += f"<p>Distance: {dist:.2f} Å ({print_pos(distv)})</p>"
            info += add_info(
                self.displayed_selection[0],
                self.displayed_structure[self.displayed_selection[0]],
            )
            info += add_info(
                self.displayed_selection[1],
                self.displayed_structure[self.displayed_selection[1]],
            )

        # Report angle geometric center and normal.
        elif len(self.displayed_selection) == 3:
            angle = self.displayed_structure.get_angle(*self.displayed_selection).round(
                2
            )
            normal = np.cross(
                *self.displayed_structure.get_distances(
                    self.displayed_selection[1],
                    [self.displayed_selection[0], self.displayed_selection[2]],
                    mic=False,
                    vector=True,
                )
            )
            normal = normal / np.linalg.norm(normal)
            info += f"<p>Angle: {angle}&deg;, Normal: ({print_pos(normal)})</p>"

        # Report dihedral angle and geometric center.
        elif len(self.displayed_selection) == 4:
            try:
                dihedral = self.displayed_structure.get_dihedral(
                    *self.displayed_selection
                )
                dihedral_str = f"{dihedral:.3f}"
            except ZeroDivisionError:
                dihedral_str = "nan"
            info += f"<p>Dihedral angle: {dihedral_str}&deg;</p>"

        # Unit and displayed cell atoms' selection.
        info_unit_and_displayed = (
            f"<p>Selected atoms: {list_to_string_range(self.displayed_selection, shift=1)}</p>"
            + f"<p>Selected unit cell atoms: {list_to_string_range(self.selection, shift=1)}</p>"
        )

        return (
            info + f"<p>Geometric center: ({geom_center})</p>" + info_unit_and_displayed
        )

    @tl.observe("displayed_selection")
    def _observe_displayed_selection_2(self, _=None):
        self.selection_info.value = self.create_selection_info()


@register_viewer_widget("data.core.folder.FolderData.")
class FolderDataViewer(ipw.VBox):
    """Viewer class for FolderData object.

    :param folder: FolderData object to be viewed
    :type folder: FolderData
    :param downloadable: If True, add link/button to download the content of the selected file in the folder
    :type downloadable: bool"""

    def __init__(self, folder, downloadable=True, **kwargs):
        self._folder = folder
        self.files = ipw.Dropdown(
            options=[obj.name for obj in self._folder.base.repository.list_objects()],
            description="Select file:",
        )
        self.text = ipw.Textarea(
            value="",
            description="File content:",
            layout={"width": "900px", "height": "300px"},
            disabled=False,
        )
        self.change_file_view()
        self.files.observe(self.change_file_view, names="value")
        children = [self.files, self.text]
        if downloadable:
            self.download_btn = ipw.Button(description="Download")
            self.download_btn.on_click(self.download)
            children.append(self.download_btn)
        super().__init__(children, **kwargs)

    def change_file_view(self, _=None):
        try:
            with self._folder.base.repository.open(self.files.value) as fobj:
                self.text.value = fobj.read()
        except UnicodeDecodeError:
            self.text.value = "[Binary file, preview not available]"

    def download(self, _=None):
        """Download selected file."""
        from IPython.display import Javascript

        # TODO: Preparing large files for download might take a while.
        # Can we do a streaming solution?
        raw_bytes = self._folder.get_object_content(self.files.value, "rb")
        base64_payload = base64.b64encode(raw_bytes).decode()

        javas = Javascript(
            f"""
            var link = document.createElement('a');
            link.href = "data:;base64,{base64_payload}"
            link.download = "{self.files.value}"
            document.body.appendChild(link);
            link.click();
            document.body.removeChild(link);
            """
        )
        display(javas)


@register_viewer_widget("data.core.array.bands.BandsData.")
class BandsDataViewer(ipw.VBox):
    """Viewer class for BandsData object.

    :param bands: BandsData object to be viewed
    :type bands: BandsData"""

    def __init__(self, bands, **kwargs):
        from bokeh.io import output_notebook, show
        from bokeh.models import Span
        from bokeh.plotting import figure

        output_notebook(hide_banner=True)
        out = ipw.Output()
        with out:
            plot_info = bands._get_bandplot_data(cartesian=True, join_symbol="|")
            # Extract relevant data
            y_data = plot_info["y"].transpose().tolist()
            x_data = [plot_info["x"] for i in range(len(y_data))]
            labels = plot_info["labels"]
            # Create the figure
            plot = figure(y_axis_label=f"Dispersion ({bands.units})")
            plot.multi_line(x_data, y_data, line_width=2, line_color="red")
            plot.xaxis.ticker = [label[0] for label in labels]
            # This trick was suggested here: https://github.com/bokeh/bokeh/issues/8166#issuecomment-426124290
            plot.xaxis.major_label_overrides = {
                int(label[0]) if label[0].is_integer() else label[0]: label[1]
                for label in labels
            }
            # Add vertical lines
            plot.renderers.extend(
                [
                    Span(
                        location=label[0],
                        dimension="height",
                        line_color="black",
                        line_width=3,
                    )
                    for label in labels
                ]
            )
            show(plot)
        children = [out]
        super().__init__(children, **kwargs)


@register_viewer_widget("process.calculation.calcfunction.CalcFunctionNode.")
@register_viewer_widget("process.calculation.calcjob.CalcJobNode.")
@register_viewer_widget("process.workflow.workfunction.WorkFunctionNode.")
@register_viewer_widget("process.workflow.workchain.WorkChainNode.")
class ProcessNodeViewerWidget(ipw.HTML):
    def __init__(self, process, **kwargs):
        self.process = process

        # Displaying reports only from the selected process,
        # NOT from its descendants.
        report = cmdline.utils.common.get_workchain_report(
            self.process, "REPORT", max_depth=1
        )
        # Filter out the first column with dates
        filtered_report = re.sub(
            r"^[0-9]{4}.*\| ([A-Z]+)\]", r"\1", report, flags=re.MULTILINE
        )
        header = f"""
            Process {process.process_label},
            State: {tools.query.formatting.format_process_state(process.process_state.value)},
            UUID: {process.uuid} (pk: {process.pk})<br>
            Started {tools.query.formatting.format_relative_time(process.ctime)},
            Last modified {tools.query.formatting.format_relative_time(process.mtime)}<br>
        """
        self.value = f"{header}<pre>{filtered_report}</pre>"

        super().__init__(**kwargs)
